from probatus.utils import (
    preprocess_data,
    shap_calc,
    calculate_shap_importance,
    BaseFitComputePlotClass,
    preprocess_labels,
    get_single_scorer,
)
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from sklearn.model_selection._search import BaseSearchCV
from sklearn.model_selection import check_cv
from sklearn.base import clone, is_classifier
from joblib import Parallel, delayed
import warnings


class ShapRFECV(BaseFitComputePlotClass):
    """
    This class performs Backwards Recursive Feature Elimination, using SHAP feature importance. At each round, for a
        given feature set, starting from all available features, the following steps are applied:

    1. (Optional) Tune the hyperparameters of the model using sklearn compatible search CV e.g.
        [GridSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LassoCV.html),
        [RandomizedSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html?highlight=randomized#sklearn.model_selection.RandomizedSearchCV), or
        [BayesSearchCV](https://scikit-optimize.github.io/stable/modules/generated/skopt.BayesSearchCV.html),
    2. Apply Cross-validation (CV) to estimate the SHAP feature importance on the provided dataset. In each CV
        iteration, the model is fitted on the train folds, and applied on the validation fold to estimate
        SHAP feature importance.
    3. Remove `step` lowest SHAP importance features from the dataset.

    At the end of the process, the user can plot the performance of the model for each iteration, and select the
        optimal number of features and the features set.

    The functionality is similar to [RFECV](https://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.RFECV.html).
        The main difference is removing the lowest importance features based on SHAP features importance. It also
        supports the use of sklearn compatible search CV for hyperparameter optimization e.g.
        [GridSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LassoCV.html),
        [RandomizedSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html?highlight=randomized#sklearn.model_selection.RandomizedSearchCV), or
        [BayesSearchCV](https://scikit-optimize.github.io/stable/modules/generated/skopt.BayesSearchCV.html), which
        needs to be passed as the `clf`. Thanks to this you can perform hyperparameter optimization at each step of
        the feature elimination. Lastly, it supports categorical features (object and category dtype) and missing values
        in the data, as long as the model supports them.

    We recommend using [LGBMClassifier](https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMClassifier.html),
        because by default it handles missing values and categorical features. In case of other models, make sure to
        handle these issues for your dataset and consider impact it might have on features importance.


    Example:
    ```python
    import numpy as np
    import pandas as pd
    from probatus.feature_elimination import ShapRFECV
    from sklearn.datasets import make_classification
    from sklearn.model_selection import train_test_split
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.model_selection import RandomizedSearchCV

    feature_names = ['f1', 'f2', 'f3', 'f4', 'f5', 'f6', 'f7', 'f8', 'f9', 'f10', 'f11', 'f12', 'f13', 'f14', 'f15', 'f16', 'f17', 'f18', 'f19', 'f20']

    # Prepare two samples
    X, y = make_classification(n_samples=200, class_sep=0.05, n_informative=6, n_features=20,
                               random_state=0, n_redundant=10, n_clusters_per_class=1)
    X = pd.DataFrame(X, columns=feature_names)


    # Prepare model and parameter search space
    clf = RandomForestClassifier(max_depth=5, class_weight='balanced')

    param_grid = {
        'n_estimators': [5, 7, 10],
        'min_samples_leaf': [3, 5, 7, 10],
    }
    search = RandomizedSearchCV(clf, param_grid)


    # Run feature elimination
    shap_elimination = ShapRFECV(
        clf=search, step=0.2, cv=10, scoring='roc_auc', n_jobs=3)
    report = shap_elimination.fit_compute(X, y)

    # Make plots
    performance_plot = shap_elimination.plot()

    # Get final feature set
    final_features_set = shap_elimination.get_reduced_features_set(num_features=3)
    ```
    <img src="../img/shaprfecv.png" width="500" />

    """

    def __init__(
        self,
        clf,
        step=1,
        min_features_to_select=1,
        cv=None,
        scoring="roc_auc",
        n_jobs=-1,
        verbose=0,
        random_state=None,
    ):
        """
        This method initializes the class:

        Args:
            clf (binary classifier, sklearn compatible search CV e.g. GridSearchCV, RandomizedSearchCV or BayesSearchCV):
                A model that will be optimized and trained at each round of features elimination. The recommended model
                is [LGBMClassifier](https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMClassifier.html),
                because it by default handles the missing values and categorical variables. This parameter also supports
                any hyperparameter search schema that is consistent with the sklearn API e.g.
                [GridSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html),
                [RandomizedSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html)
                or [BayesSearchCV](https://scikit-optimize.github.io/stable/modules/generated/skopt.BayesSearchCV.html#skopt.BayesSearchCV).

            step (int or float, optional):
                Number of lowest importance features removed each round. If it is an int, then each round such number of
                features is discarded. If float, such percentage of remaining features (rounded down) is removed each
                iteration. It is recommended to use float, since it is faster for a large number of features, and slows
                down and becomes more precise towards less features. Note: the last round may remove fewer features in
                order to reach min_features_to_select.
                If columns_to_keep parameter is specified in the fit method, step is the number of features to remove after
                keeping those columns.

            min_features_to_select (int, optional):
                Minimum number of features to be kept. This is a stopping criterion of the feature elimination. By
                default the process stops when one feature is left. If columns_to_keep is specified in the fit method,
                it may overide this parameter to the maximum between length of columns_to_keep the two.

            cv (int, cross-validation generator or an iterable, optional):
                Determines the cross-validation splitting strategy. Compatible with sklearn
                [cv parameter](https://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.RFECV.html).
                If None, then cv of 5 is used.

            scoring (string or probatus.utils.Scorer, optional):
                Metric for which the model performance is calculated. It can be either a metric name  aligned with
                predefined [classification scorers names in sklearn](https://scikit-learn.org/stable/modules/model_evaluation.html).
                Another option is using probatus.utils.Scorer to define a custom metric.

            n_jobs (int, optional):
                Number of cores to run in parallel while fitting across folds. None means 1 unless in a
                `joblib.parallel_backend` context. -1 means using all processors.

            verbose (int, optional):
                Controls verbosity of the output:

                - 0 - nether prints nor warnings are shown
                - 1 - 50 - only most important warnings
                - 51 - 100 - shows other warnings and prints
                - above 100 - presents all prints and all warnings (including SHAP warnings).

            random_state (int, optional):
                Random state set at each round of feature elimination. If it is None, the results will not be
                reproducible and in random search at each iteration a different hyperparameters might be tested. For
                reproducible results set it to integer.
        """
        self.clf = clf

        if isinstance(self.clf, BaseSearchCV):
            self.search_clf = True
        else:
            self.search_clf = False

        if (isinstance(step, int) or isinstance(step, float)) and step > 0:
            self.step = step
        else:
            raise (
                ValueError(
                    f"The current value of step = {step} is not allowed. "
                    f"It needs to be a positive integer or positive float."
                )
            )

        if isinstance(min_features_to_select, int) and min_features_to_select > 0:
            self.min_features_to_select = min_features_to_select
        else:
            raise (
                ValueError(
                    f"The current value of min_features_to_select = {min_features_to_select} is not allowed. "
                    f"It needs to be a greater than or equal to 0."
                )
            )

        self.cv = cv
        self.scorer = get_single_scorer(scoring)
        self.random_state = random_state
        self.n_jobs = n_jobs
        self.report_df = pd.DataFrame([])
        self.verbose = verbose

    def _get_current_features_to_remove(self, shap_importance_df, columns_to_keep=None):
        """
        Implements the logic used to determine which features to remove. If step is a positive integer,
            at each round step lowest SHAP importance features are selected. If it is a float, such percentage
            of remaining features (rounded up) is removed each iteration. It is recommended to use float, since it is
            faster for a large set of features, and slows down and becomes more precise towards less features.

        Args:
            shap_importance_df (pd.DataFrame):
                DataFrame presenting SHAP importance of remaining features.

        Returns:
            (list):
                List of features to be removed at a given round.
        """

        # Bounding the variable.
        num_features_to_remove = 0

        # If columns_to_keep is not None, exclude those columns and
        # calculate features to remove.
        if columns_to_keep is not None:
            mask = shap_importance_df.index.isin(columns_to_keep)
            shap_importance_df = shap_importance_df[~mask]

        # If the step is an int remove n features.
        if isinstance(self.step, int):
            num_features_to_remove = self._calculate_number_of_features_to_remove(
                current_num_of_features=shap_importance_df.shape[0],
                num_features_to_remove=self.step,
                min_num_features_to_keep=self.min_features_to_select,
            )
        # If the step is a float remove n * number features that are left, rounded down
        elif isinstance(self.step, float):
            current_step = int(np.floor(shap_importance_df.shape[0] * self.step))
            # The step after rounding down should be at least 1
            if current_step < 1:
                current_step = 1

            num_features_to_remove = self._calculate_number_of_features_to_remove(
                current_num_of_features=shap_importance_df.shape[0],
                num_features_to_remove=current_step,
                min_num_features_to_keep=self.min_features_to_select,
            )

        if num_features_to_remove == 0:
            return []
        else:
            return shap_importance_df.iloc[-num_features_to_remove:].index.tolist()

    @staticmethod
    def _calculate_number_of_features_to_remove(
        current_num_of_features, num_features_to_remove, min_num_features_to_keep
    ):
        """
        Calculates the number of features to be removed, and makes sure that after removal at least
            min_num_features_to_keep are kept

         Args:
            current_num_of_features (int):
                Current number of features in the data.

            num_features_to_remove (int):
                Number of features to be removed at this stage.

            min_num_features_to_keep (int):
                Minimum number of features to be left after removal.

        Returns:
            (int):
                Number of features to be removed.
        """
        num_features_after_removal = current_num_of_features - num_features_to_remove
        if num_features_after_removal >= min_num_features_to_keep:
            num_to_remove = num_features_to_remove
        else:
            # take all available features minus number of them that should stay
            num_to_remove = current_num_of_features - min_num_features_to_keep
        return num_to_remove

    def _report_current_results(
        self,
        round_number,
        current_features_set,
        features_to_remove,
        train_metric_mean,
        train_metric_std,
        val_metric_mean,
        val_metric_std,
    ):
        """
        This function adds the results from a current iteration to the report.

        Args:
            round_number (int):
                Current number of the round.

            current_features_set (list of str):
                Current list of features.

            features_to_remove (list of str):
                List of features to be removed at the end of this iteration.

            train_metric_mean (float or int):
                Mean scoring metric measured on train set during CV.

            train_metric_std (float or int):
                Std scoring metric measured on train set during CV.

            val_metric_mean (float or int):
                Mean scoring metric measured on validation set during CV.

            val_metric_std (float or int):
                Std scoring metric measured on validation set during CV.
        """

        current_results = {
            "num_features": len(current_features_set),
            "features_set": None,
            "eliminated_features": None,
            "train_metric_mean": train_metric_mean,
            "train_metric_std": train_metric_std,
            "val_metric_mean": val_metric_mean,
            "val_metric_std": val_metric_std,
        }

        current_row = pd.DataFrame(current_results, index=[round_number])
        current_row["features_set"] = [current_features_set]
        current_row["eliminated_features"] = [features_to_remove]

        self.report_df = pd.concat([self.report_df, current_row], axis=0)

    @staticmethod
    def _get_feature_shap_values_per_fold(
        X, y, clf, train_index, val_index, scorer, verbose=0
    ):
        """
        This function calculates the shap values on validation set, and Train and Val score.

        Args:
            X (pd.DataFrame):
                Dataset used in CV.

            y (pd.Series):
                Binary labels for X.

            clf (binary classifier):
                Model to be fitted on the train folds.

            train_index (np.array):
                Positions of train folds samples.

            val_index (np.array):
                Positions of validation fold samples.

            scorer (string, callable or None):
                A string (see sklearn [model scoring](https://scikit-learn.org/stable/modules/model_evaluation.html)) or
                a scorer callable object, function with the signature `scorer(estimator, X, y)`.

            verbose (int, optional):
                Controls verbosity of the output:

                - 0 - neither prints nor warnings are shown
                - 1 - 50 - only most important warnings regarding data properties are shown (excluding SHAP warnings)
                - 51 - 100 - shows most important warnings, prints of the feature removal process
                - above 100 - presents all prints and all warnings (including SHAP warnings).

        Returns:
            (np.array, float, float):
                Tuple with the results: Shap Values on validation fold, train score, validation score.
        """
        X_train, X_val = X.iloc[train_index, :], X.iloc[val_index, :]
        y_train, y_val = y.iloc[train_index], y.iloc[val_index]

        # Fit model with train folds
        clf = clf.fit(X_train, y_train)

        # Score the model
        score_train = scorer(clf, X_train, y_train)
        score_val = scorer(clf, X_val, y_val)

        # Compute SHAP values
        shap_values = shap_calc(clf, X_val, verbose=verbose)
        return shap_values, score_train, score_val

    def fit(self, X, y, columns_to_keep=None, column_names=None):
        """
        Fits the object with the provided data. The algorithm starts with the entire dataset, and then sequentially
             eliminates features. If sklearn compatible search CV is passed as clf e.g.
             [GridSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html),
             [RandomizedSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html)
             or [BayesSearchCV](https://scikit-optimize.github.io/stable/modules/generated/skopt.BayesSearchCV.html),
             the hyperparameter optimization is applied at each step of the elimination. Then, the SHAP feature importance
             is calculated using Cross-Validation, and `step` lowest importance features are removed.

        Args:
            X (pd.DataFrame):
                Provided dataset.

            y (pd.Series):
                Binary labels for X.

            columns_to_keep (list of str, optional):
                List of column names to keep. If given, these columns will not be eliminated by the feature elimination process.
                However, these feature will used for the calculation of the SHAP values.

            column_names (list of str, optional):
                List of feature names of the provided samples. If provided it will be used to overwrite the existing
                feature names. If not provided the existing feature names are used or default feature names are
                generated.
        Returns:
            (ShapRFECV): Fitted object.
        """
        # Set seed for results reproducibility
        if self.random_state is not None:
            np.random.seed(self.random_state)

        # If to columns_to_keep is not provided, then initialise it by an empty string.
        # If provided check if all the elements in columns_to_keep are of type string.
        if columns_to_keep is None:
            len_columns_to_keep = 0
        else:
            if all(isinstance(x, str) for x in columns_to_keep):
                len_columns_to_keep = len(columns_to_keep)
            else:
                raise (
                    ValueError(
                        "The current values of columns_to_keep are not allowed.All the elements should be strings."
                    )
                )

        # If the columns_to_keep parameter is provided, check if they match the column names in the X.
        if column_names is not None:
            if all(x in column_names for x in list(X.columns)):
                pass
            else:
                raise (
                    ValueError(
                        "The column names in parameter columns_to_keep and column_names are not macthing."
                    )
                )

        # Check that the total number of columns to select is less than total number of columns in the data.
        # only when both parameters are provided.
        if column_names is not None and columns_to_keep is not None:
            if (self.min_features_to_select + len_columns_to_keep) > len(
                self.column_names
            ):
                raise ValueError(
                    "Minimum features to select is greater than number of features."
                    "Lower the value for min_features_to_select or number of columns in columns_to_keep"
                )

        self.X, self.column_names = preprocess_data(
            X, X_name="X", column_names=column_names, verbose=self.verbose
        )
        self.y = preprocess_labels(
            y, y_name="y", index=self.X.index, verbose=self.verbose
        )
        self.cv = check_cv(self.cv, self.y, classifier=is_classifier(self.clf))

        remaining_features = current_features_set = self.column_names
        round_number = 0

        # Stop when stopping criteria is met.
        stopping_criteria = np.max([self.min_features_to_select, len_columns_to_keep])

        # Setting up the min_features_to_select parameter.
        if columns_to_keep is None:
            pass
        else:
            self.min_features_to_select = 0
            # This ensures that, if columns_to_keep is provided ,the last features remaining are only the columns_to_keep.
            if self.verbose > 50:
                warnings.warn(f"Minimum features to select : {stopping_criteria}")

        while len(current_features_set) > stopping_criteria:
            round_number += 1

            # Get current dataset info
            current_features_set = remaining_features
            if columns_to_keep is None:
                remaining_removeable_features = list(set(current_features_set))
            else:
                remaining_removeable_features = list(
                    set(current_features_set) | set(columns_to_keep)
                )
            current_X = self.X[remaining_removeable_features]

            # Set seed for results reproducibility
            if self.random_state is not None:
                np.random.seed(self.random_state)

            # Optimize parameters
            if self.search_clf:
                current_search_clf = clone(self.clf).fit(current_X, self.y)
                current_clf = current_search_clf.estimator.set_params(
                    **current_search_clf.best_params_
                )
            else:
                current_clf = clone(self.clf)

            # Perform CV to estimate feature importance with SHAP
            results_per_fold = Parallel(n_jobs=self.n_jobs)(
                delayed(self._get_feature_shap_values_per_fold)(
                    X=current_X,
                    y=self.y,
                    clf=current_clf,
                    train_index=train_index,
                    val_index=val_index,
                    scorer=self.scorer.scorer,
                    verbose=self.verbose,
                )
                for train_index, val_index in self.cv.split(current_X, self.y)
            )

            shap_values = np.vstack(
                [current_result[0] for current_result in results_per_fold]
            )
            scores_train = [current_result[1] for current_result in results_per_fold]
            scores_val = [current_result[2] for current_result in results_per_fold]

            # Calculate the shap features with remaining features and features to keep.

            shap_importance_df = calculate_shap_importance(
                shap_values, remaining_removeable_features
            )

            # Get features to remove
            features_to_remove = self._get_current_features_to_remove(
                shap_importance_df, columns_to_keep=columns_to_keep
            )
            remaining_features = list(
                set(current_features_set) - set(features_to_remove)
            )

            # Report results
            self._report_current_results(
                round_number=round_number,
                current_features_set=current_features_set,
                features_to_remove=features_to_remove,
                train_metric_mean=np.round(np.mean(scores_train), 3),
                train_metric_std=np.round(np.std(scores_train), 3),
                val_metric_mean=np.round(np.mean(scores_val), 3),
                val_metric_std=np.round(np.std(scores_val), 3),
            )
            if self.verbose > 50:
                print(
                    f"Round: {round_number}, Current number of features: {len(current_features_set)}, "
                    f'Current performance: Train {self.report_df.loc[round_number]["train_metric_mean"]} '
                    f'+/- {self.report_df.loc[round_number]["train_metric_std"]}, CV Validation '
                    f'{self.report_df.loc[round_number]["val_metric_mean"]} '
                    f'+/- {self.report_df.loc[round_number]["val_metric_std"]}. \n'
                    f"Features left: {remaining_features}. "
                    f"Removed features at the end of the round: {features_to_remove}"
                )
        self.fitted = True
        return self

    def compute(self):
        """
        Checks if fit() method has been run and computes the DataFrame with results of feature elimintation for each
         round.

        Returns:
            (pd.DataFrame):
                DataFrame with results of feature elimination for each round.
        """
        self._check_if_fitted()

        return self.report_df

    def fit_compute(self, X, y, columns_to_keep=None, column_names=None):
        """
        Fits the object with the provided data. The algorithm starts with the entire dataset, and then sequentially
             eliminates features. If sklearn compatible search CV is passed as clf e.g.
             [GridSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html),
             [RandomizedSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html)
             or [BayesSearchCV](https://scikit-optimize.github.io/stable/modules/generated/skopt.BayesSearchCV.html),
             the hyperparameter optimization is applied at each step of the elimination. Then, the SHAP feature importance
             is calculated using Cross-Validation, and `step` lowest importance features are removed. At the end, the
             report containing results from each iteration is computed and returned to the user.

        Args:
            X (pd.DataFrame):
                Provided dataset.

            y (pd.Series):
                Binary labels for X.

            columns_to_keep (list of str, optional):
                List of columns to keep. If given, these columns will not be eliminated.

            column_names (list of str, optional):
                List of feature names of the provided samples. If provided it will be used to overwrite the existing
                feature names. If not provided the existing feature names are used or default feature names are
                generated.

        Returns:
            (pd.DataFrame):
                DataFrame containing results of feature elimination from each iteration.
        """

        self.fit(X, y, columns_to_keep=columns_to_keep, column_names=column_names)
        return self.compute()

    def get_reduced_features_set(self, num_features):
        """
        Gets the features set after the feature elimination process, for a given number of features.

        Args:
            num_features (int):
                Number of features in the reduced features set.

        Returns:
            (list of str):
                Reduced features set.
        """
        self._check_if_fitted()

        if num_features not in self.report_df.num_features.tolist():
            raise (
                ValueError(
                    f"The provided number of features has not been achieved at any stage of the process. "
                    f"You can select one of the following: {self.report_df.num_features.tolist()}"
                )
            )
        else:
            return self.report_df[self.report_df.num_features == num_features][
                "features_set"
            ].values[0]

    def plot(self, show=True, **figure_kwargs):
        """
        Generates plot of the model performance for each iteration of feature elimination.

        Args:
            show (bool, optional):
                If True, the plots are showed to the user, otherwise they are not shown. Not showing plot can be useful,
                when you want to edit the returned axis, before showing it.

            **figure_kwargs:
                Keyword arguments that are passed to the plt.figure, at its initialization.

        Returns:
            (plt.axis):
                Axis containing the performance plot.
        """
        x_ticks = list(reversed(self.report_df["num_features"].tolist()))

        plt.figure(**figure_kwargs)

        plt.plot(
            self.report_df["num_features"],
            self.report_df["train_metric_mean"],
            label="Train Score",
        )
        plt.fill_between(
            pd.to_numeric(self.report_df.num_features, errors="coerce"),
            self.report_df["train_metric_mean"] - self.report_df["train_metric_std"],
            self.report_df["train_metric_mean"] + self.report_df["train_metric_std"],
            alpha=0.3,
        )

        plt.plot(
            self.report_df["num_features"],
            self.report_df["val_metric_mean"],
            label="Validation Score",
        )
        plt.fill_between(
            pd.to_numeric(self.report_df.num_features, errors="coerce"),
            self.report_df["val_metric_mean"] - self.report_df["val_metric_std"],
            self.report_df["val_metric_mean"] + self.report_df["val_metric_std"],
            alpha=0.3,
        )

        plt.xlabel("Number of features")
        plt.ylabel(f"Performance {self.scorer.metric_name}")
        plt.title("Backwards Feature Elimination using SHAP & CV")
        plt.legend(loc="lower left")
        ax = plt.gca()
        ax.invert_xaxis()
        ax.set_xticks(x_ticks)
        if show:
            plt.show()
        else:
            plt.close()
        return ax



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import numpy as np
import pandas as pd


def sample_data(X, y, sampling_type, sampling_fraction, dataset_name="dataset"):
    check_sampling_input(sampling_type, sampling_fraction, dataset_name)

    if sampling_type is None:
        return X, y

    number_of_samples = np.ceil(sampling_fraction * X.shape[0]).astype(int)
    array_index = list(range(X.shape[0]))

    if sampling_type == "bootstrap":
        rows_indexes = np.random.choice(array_index, number_of_samples, replace=True)
    else:
        if sampling_fraction == 1 or number_of_samples == X.shape[0]:
            return X, y
        else:
            rows_indexes = np.random.choice(
                array_index, number_of_samples, replace=True
            )

    # Get output correctly based on the type
    if isinstance(X, pd.DataFrame):
        output_X = X.iloc[rows_indexes]
    else:
        output_X = X[rows_indexes]
    if isinstance(y, pd.DataFrame):
        output_y = y.iloc[rows_indexes]
    else:
        output_y = y[rows_indexes]

    return output_X, output_y


def check_sampling_input(sampling_type, fraction, dataset_name):
    if sampling_type is not None:
        if sampling_type == "bootstrap":
            if fraction <= 0:
                raise (
                    ValueError(
                        f"For bootstrapping {dataset_name} fraction needs to be above 0"
                    )
                )
        elif sampling_type == "subsample":
            if fraction <= 0 or fraction >= 1:
                raise (
                    ValueError(
                        f"For bootstrapping {dataset_name} fraction needs to be be above 0 and below 1"
                    )
                )
        else:
            raise (ValueError("This sampling method is not implemented"))



import logging

import numpy as np
import pandas as pd
import pytest

from probatus.utils.sampling import sample_row

LOGGER = logging.getLogger(__name__)


@pytest.fixture(scope="function")
def test_dataframe():
    """Returns a simple dataframe to test on"""
    return pd.DataFrame({"col1": [0, 1, 2, 3], "col2": ["a", "b", "c", "d"],})


@pytest.fixture(scope="function")
def expected_test_return():
    """Expected return for test_dataframe"""
    return pd.DataFrame(
        {
            "column": ["col1", "col2"],
            "dtype": [np.dtype("int64"), np.dtype("O")],
            "sample": [1, "b"],
            "range_low": [0, ""],
            "range_high": [3, ""],
        }
    ).set_index(["column"])


@pytest.fixture(scope="function")
def empty_dataframe():
    """Returns an empty dataframe to test on"""
    return pd.DataFrame([])


@pytest.fixture(scope="function")
def na_col_dataframe():
    """Returns a dataframe with a completely NaN column to test on"""
    return pd.DataFrame(
        {"col1": [0, 1, 2, 3], "col2": ["a", np.nan, "c", "d"], "colNaN": [np.nan] * 4}
    )


@pytest.fixture(scope="function")
def na_in_row_dataframe():
    """Returns a dataframe with a completely NaN column to test on"""
    return pd.DataFrame({"col1": [0, 1, 2, 3], "col2": ["a", np.nan, "c", "d"]})


@pytest.fixture(scope="function")
def expected_na_in_row_return():
    """Expected return for test_dataframe"""
    return pd.DataFrame(
        {
            "column": ["col1", "col2"],
            "dtype": [np.dtype("int64"), np.dtype("O")],
            "sample": [0, "a"],
            "range_low": [0, ""],
            "range_high": [3, ""],
        },
    ).set_index(["column"])


@pytest.fixture(scope="function")
def expected_na_col_return_w_filtering():
    """Expected return for test_dataframe"""
    return pd.DataFrame(
        {
            "column": ["col1", "col2", "colNaN"],
            "dtype": [np.dtype("int64"), np.dtype("O"), np.dtype("float64")],
            "sample": [1, np.nan, np.nan],
            "range_low": [0, "", np.nan],
            "range_high": [3, "", np.nan],
        }
    ).set_index(["column"])


@pytest.fixture(scope="function")
def long_str_dataframe():
    """Returns a DataFrame with a long string to test truncation"""
    return pd.DataFrame(
        {"col1": ["This is a very long string, hello, how nice of you to read this!"]}
    )


@pytest.fixture(scope="function")
def expected_long_str_return():
    """Expected return for long_str_dataframe"""
    return pd.DataFrame(
        {
            "column": ["col1"],
            "dtype": [np.dtype("O")],
            "sample": ["This is a very long stri...ce of you to read this!"],
            "range_low": [""],
            "range_high": [""],
        }
    ).set_index(["column"])


def test_normal_df(test_dataframe, expected_test_return):
    pd.testing.assert_frame_equal(sample_row(test_dataframe), expected_test_return)


def test_empty_df(empty_dataframe):
    with pytest.raises(AssertionError):
        sample_row(empty_dataframe)


def test_df_with_nan_column(
    na_col_dataframe, expected_na_col_return_w_filtering, caplog
):
    with caplog.at_level("INFO"):
        pd.testing.assert_frame_equal(
            sample_row(na_col_dataframe, random_state=42, filter_rows_with_na=True),
            expected_na_col_return_w_filtering,
        )

    assert "No rows without NaN found, sampling from all rows.." in caplog.text


def test_df_with_nan_in_row(na_in_row_dataframe, expected_na_in_row_return):
    pd.testing.assert_frame_equal(
        sample_row(na_in_row_dataframe, random_state=42, filter_rows_with_na=True),
        expected_na_in_row_return,
    )


def test_long_str_df(long_str_dataframe, expected_long_str_return):
    pd.testing.assert_frame_equal(
        sample_row(long_str_dataframe, random_state=42), expected_long_str_return
    )


def test_filter_rows_with_na_parameter(na_col_dataframe):
    with pytest.raises(AssertionError):
        sample_row(test_dataframe, filter_rows_with_na="a")
    with pytest.raises(AssertionError):
        sample_row(test_dataframe, filter_rows_with_na=None)


def test_random_state_parameter(test_dataframe):
    with pytest.raises(AssertionError):
        sample_row(test_dataframe, random_state="a")
    with pytest.raises(AssertionError):
        sample_row(test_dataframe, random_state=None)


def test_max_field_len_parameter(test_dataframe):
    with pytest.raises(AssertionError):
        sample_row(test_dataframe, random_state="a")
    with pytest.raises(AssertionError):
        sample_row(test_dataframe, random_state=None)



# This approach is adapted from, and explained in: https://calmcode.io/docs/epic.html

import pytest
import os

# Turn off interactive mode in plots

plots_disable = "import matplotlib \n" \
                "import matplotlib.pyplot as plt \n" \
                "plt.ioff() \n" \
                "matplotlib.use('Agg') \n" \

NOTEBOOKS_PATH = './docs/tutorials/'

NOTEBOOKS_TO_TEST_LGBM = [
    'nb_shap_feature_elimination',
]

NOTEBOOKS_TO_TEST = [
    'nb_binning',
    'nb_custom_scoring',
    'nb_distribution_statistics',
    'nb_metric_volatility',
    'nb_sample_similarity',
    'nb_shap_model_interpreter'
]


def get_notebook_path(notebook_name):
    return os.path.join(NOTEBOOKS_PATH, notebook_name + '.ipynb')


def execute_notebook_test(notebook_name):
    notebook_path = get_notebook_path(notebook_name)
    code_to_execute = os.popen(f"jupyter nbconvert --to script --execute --stdout {notebook_path}").read()
    _ = os.popen(f"python3 -c {plots_disable + code_to_execute}").read()


@pytest.mark.parametrize("notebook_name", NOTEBOOKS_TO_TEST_LGBM)
@pytest.mark.skipif(os.environ.get("SKIP_LIGHTGBM") == 'true', reason="LightGBM tests disabled")
def test_jupyter_notebook_lgbm(notebook_name):
    execute_notebook_test(notebook_name)


@pytest.mark.parametrize("notebook_name", NOTEBOOKS_TO_TEST)
def test_jupyter_notebook_lgbm(notebook_name):
    execute_notebook_test(notebook_name)



"""
Utility functions for sampling

This module holds utility functions for sampling data. They can be used, for
example, to provide a random single typical datapoin in a technical report.
"""

import logging

import pandas as pd
from pandas.api.types import is_numeric_dtype


def sample_row(
    X: pd.DataFrame,
    filter_rows_with_na: bool = False,
    random_state: int = 42,
    max_field_len: int = 50,
) -> pd.DataFrame:
    """Sample a row from pandas dataframe.

    Extracts the column name, datatype, minimum and maximum values for each
    column in the supplied dataframe. The orientation is row-based (as opposed to `df.sample(1)`), which allows
    for better printing when a dataset contains many features. This function is usefull when providing a
    sample row in technical model documentation.

    Example:

    ```python
    from probatus.utils import sample_row
    from sklearn.datasets import load_iris

    iris = load_iris(as_frame=True).get('data')
    sample = sample_row(iris, filter_rows_with_na=False, random_state=12)
    print(sample.to_markdown())
    ```

    ??? info "Example output"

        | column            | dtype   |   sample |   range_low |   range_high |
        |:------------------|:--------|---------:|------------:|-------------:|
        | sepal length (cm) | float64 |      5   |         4.3 |          7.9 |
        | sepal width (cm)  | float64 |      3.5 |         2   |          4.4 |
        | petal length (cm) | float64 |      1.3 |         1   |          6.9 |
        | petal width (cm)  | float64 |      0.3 |         0.1 |          2.5 |

    Args:
        X (DataFrame):
            Pandas DataFrame to be sampled
        filter_rows_with_na (bool):
            if true, rows with na values are not considered for sampling
        random_state (int):
            Optional random state to ensure reproducability
        max_field_len (int):
            Maximum number of characters for fields, beyond
            which any text is truncated

    Returns:
        (pd.DataFrame): A Pandas DataFrame containing the sampled row
    """
    # Input validation
    assert type(X) == pd.DataFrame, "X should be pandas DataFrame"
    assert X.empty is False, "X should not be an empty DataFrame"
    assert type(filter_rows_with_na) == bool, "filter_rows_with_na should be a boolean"
    assert type(random_state) == int, "random_state should be an integer"
    assert type(max_field_len) == int, "max_field_len should be an integer"

    # Create new empty df
    sample_df = pd.DataFrame()

    # Convert dtypes of pandas to ensure detection of data types
    X_dtypes = X.convert_dtypes()

    # Sample row from X
    sample_row = X.sample(1, random_state=random_state)
    if filter_rows_with_na:
        try:
            sample_row = X.dropna().sample(1, random_state=random_state)
        except ValueError:
            logging.info(
                "sample_row(): No rows without NaN found, sampling from all rows.."
            )

    # Sample every column of X
    for i, col in enumerate(sample_row.columns):
        # Extract sample from X if not all samples are nan
        sample = sample_row[col].values[0]

        # If datatype allows it, extract low and high range
        if is_numeric_dtype(X_dtypes[col]):
            low = X[col].min(skipna=True)
            high = X[col].max(skipna=True)
        else:
            low = ""
            high = ""

            # Shorten sampled datapoint if too long
            if isinstance(sample, str) and len(sample) > max_field_len:
                sample = (
                    sample[: (max_field_len // 2) - 1]
                    + "..."
                    + sample[(-max_field_len // 2) + 2 :]
                )

        # Add new row to sample_df
        row_df = pd.DataFrame(
            {
                "column": [col],
                "dtype": [X[col].dtype],
                "sample": [sample],
                "range_low": [low],
                "range_high": [high],
            }
        )
        sample_df = pd.concat([sample_df, row_df], ignore_index=True)

    sample_df = sample_df.set_index(["column"])
    return sample_df



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import shap
import pandas as pd
import numpy as np
import warnings


def shap_calc(
    model,
    X,
    approximate=False,
    return_explainer=False,
    verbose=0,
    sample_size=100,
    **shap_kwargs,
):
    """
    Helper function to calculate the shapley values for a given model.

    Args:
        model (binary model):
            Trained model.

        X (pd.DataFrame or np.ndarray):
            features set.

         approximate (boolean):
            if True uses shap approximations - less accurate, but very fast.

        return_explainer (boolean):
            if True, returns a a tuple (shap_values, explainer).

        verbose (int, optional):
            Controls verbosity of the output:

            - 0 - nether prints nor warnings are shown
            - 1 - 50 - only most important warnings
            - 51 - 100 - shows other warnings and prints
            - above 100 - presents all prints and all warnings (including SHAP warnings).

        **shap_kwargs: kwargs of the shap.Explainer

    Returns:
        (np.ndarray or tuple(np.ndarray, shap.Explainer)):
            shapley_values for the model, optionally also returns the explainer.

    """
    # Suppress warnings regarding XGboost and Lightgbm models.
    with warnings.catch_warnings():
        if verbose <= 100:
            warnings.simplefilter("ignore")

        # Create the background data,required for non tree based models.
        # A single datapoint can passed as mask (https://github.com/slundberg/shap/issues/955#issuecomment-569837201)

        if X.shape[1] < sample_size:
            sample_size = int(np.ceil(X.shape[1] * 0.2))
        else:
            pass
        mask = shap.utils.sample(X, sample_size)

        explainer = shap.Explainer(model, masker=mask, **shap_kwargs)
        # Calculate Shap values.
        shap_values = explainer.shap_values(X)

        if isinstance(shap_values, list) and len(shap_values) == 2:
            warnings.warn(
                "Shap values are related to the output probabilities of class 1 for this model, instead of "
                "log odds."
            )
            shap_values = shap_values[1]

    if return_explainer:
        return shap_values, explainer
    return shap_values


def shap_to_df(model, X, precalc_shap=None, **kwargs):
    """
    Calculates the shap values and return the pandas DataFrame with the columns and the index of the original.

    Args:
        model (binary model):
            Pretrained model (Random Forest of XGBoost at the moment).

        X (pd.DataFrame or np.ndarray):
            Dataset on which the SHAP importance is calculated.

        precalc_shap (np.array):
            Precalculated SHAP values. If None, they are computed.

        **kwargs: for the function shap_calc

    Returns:
        (pd.DataFrame):
            Dataframe with SHAP feature importance per features on X dataset.
    """
    if precalc_shap is not None:
        shap_values = precalc_shap
    else:
        shap_values = shap_calc(model, X, **kwargs)
    if isinstance(X, pd.DataFrame):
        return pd.DataFrame(shap_values, columns=X.columns, index=X.index)

    elif isinstance(X, np.ndarray) and len(X.shape) == 2:
        return pd.DataFrame(
            shap_values, columns=[f"col_{ix}" for ix in range(X.shape[1])]
        )

    else:
        raise NotImplementedError("X must be a dataframe or a 2d array")


def calculate_shap_importance(shap_values, columns, output_columns_suffix=""):
    """
    Returns the average shapley value for each column of the dataframe, as well as the average absolute shap value.

    Args:
        shap_values (np.array):
            Shap values.

        columns (list of str):
            Feature names.

        output_columns_suffix (str, optional):
            Suffix to be added at the end of column names in the output.

    Returns:
        (pd.DataFrame):
            Mean absolute shap values and Mean shap values of features.

    """

    # Find average shap importance for neg and pos class
    shap_abs_mean = np.mean(np.abs(shap_values), axis=0)
    shap_mean = np.mean(shap_values, axis=0)

    # Prepare importance values in a handy df
    importance_df = pd.DataFrame(
        {
            f"mean_abs_shap_value{output_columns_suffix}": shap_abs_mean.tolist(),
            f"mean_shap_value{output_columns_suffix}": shap_mean.tolist(),
        },
        index=columns,
    )

    # Set the correct column types
    importance_df[f"mean_abs_shap_value{output_columns_suffix}"] = importance_df[
        f"mean_abs_shap_value{output_columns_suffix}"
    ].astype(float)
    importance_df[f"mean_shap_value{output_columns_suffix}"] = importance_df[
        f"mean_shap_value{output_columns_suffix}"
    ].astype(float)

    importance_df = importance_df.sort_values(
        f"mean_abs_shap_value{output_columns_suffix}", ascending=False
    )

    return importance_df



import numpy as np
import pandas as pd

import pytest

from probatus.utils import assure_numpy_array, assure_pandas_df, check_1d, DimensionalityError, check_numeric_dtypes, \
    preprocess_labels, preprocess_data


def test_assure_numpy_array_list():
    x = [1, 2, 3]
    x_array = assure_numpy_array(x)
    assert isinstance(x_array, np.ndarray)
    np.testing.assert_array_equal(x_array, np.array(x))
    x = [[1, 2], [3, 4]]
    x_array = assure_numpy_array(x)
    np.testing.assert_array_equal(x_array, np.array([[1, 2], [3, 4]]))
    with pytest.raises(DimensionalityError):
        assert assure_numpy_array(x, assure_1d=True)


def test_assure_numpy_array_array():
    x = np.array([1, 2, 3])
    x_array = assure_numpy_array(x)
    assert isinstance(x_array, np.ndarray)
    np.testing.assert_array_equal(x_array, x)
    x = np.array([[1, 2], [3, 4]])
    x_array = assure_numpy_array(x)
    np.testing.assert_array_equal(x_array, x)
    with pytest.raises(DimensionalityError):
        assert assure_numpy_array(x, assure_1d=True)


def test_assure_numpy_array_dataframe():
    x = pd.DataFrame({'x': [1, 2, 3]})
    x_array = assure_numpy_array(x)
    assert isinstance(x_array, np.ndarray)
    np.testing.assert_array_equal(x_array, np.array([1, 2, 3]))
    x = pd.DataFrame({'x': [1, 2, 3], 'y': [1, 2, 3]})
    x_array = assure_numpy_array(x)
    np.testing.assert_array_equal(x_array, np.array([[1, 1], [2, 2], [3, 3]]))
    with pytest.raises(DimensionalityError):
        assert assure_numpy_array(x, assure_1d=True)


def test_assure_numpy_array_series():
    x = pd.Series([1, 2, 3])
    x_array = assure_numpy_array(x)
    assert isinstance(x_array, np.ndarray)
    np.testing.assert_array_equal(x_array, np.array([1, 2, 3]))


def test_check_1d_list():
    x = [1, 2, 3]
    assert check_1d(x)
    y = [[1, 2], [1, 2, 3]]
    with pytest.raises(DimensionalityError):
        assert check_1d(y)
    y = [1, [1, 2, 3]]
    with pytest.raises(DimensionalityError):
        assert check_1d(y)


def test_check_1d_array():
    x = np.array([1, 2, 3])
    assert check_1d(x)
    y = np.array([[1, 2], [1, 2, 3]])
    with pytest.raises(DimensionalityError):
        assert check_1d(y)
    y = np.array([0, [1, 2, 3]])
    with pytest.raises(DimensionalityError):
        assert check_1d(y)


def test_check_1d_dataframe():
    x = pd.DataFrame({'x': [1, 2, 3]})
    assert check_1d(x)
    y = pd.DataFrame({'x': [1, 2, 3], 'y': [1, 2, 3]})
    with pytest.raises(DimensionalityError):
        assert check_1d(y)
    y = pd.DataFrame({'x': [1, 2, 3, [4, 5]]})
    with pytest.raises(DimensionalityError):
        assert check_1d(y)


def test_check_1d_series():
    x = pd.Series([1, 2, 3])
    assert check_1d(x)
    y = pd.Series([1, [2, 3]])
    with pytest.raises(DimensionalityError):
        assert check_1d(y)
    y = pd.Series([[1], [2, 3]])
    with pytest.raises(DimensionalityError):
        assert check_1d(y)

@pytest.fixture(scope="function")
def expected_df_2d():
    return pd.DataFrame({0: [1, 2], 1: [2, 3], 2: [3, 4]})

@pytest.fixture(scope="function")
def expected_df():
    return pd.DataFrame({0: [1, 2, 3]})

def test_assure_pandas_df_list(expected_df):
    x = [1, 2, 3]
    x_df = assure_pandas_df(x)
    assert x_df.equals(expected_df)

def test_assure_pandas_df_list_of_lists(expected_df_2d):
    x = [[1, 2, 3], [2, 3, 4]]
    x_df = assure_pandas_df(x)
    assert x_df.equals(expected_df_2d)
    
def test_assure_pandas_df_series(expected_df):
    x = pd.Series([1, 2, 3])
    x_df = assure_pandas_df(x)
    assert x_df.equals(expected_df)
    
def test_assure_pandas_df_array(expected_df, expected_df_2d):
    x = np.array([[1, 2, 3], [2, 3, 4]])
    x_df = assure_pandas_df(x)
    assert x_df.equals(expected_df_2d)
    
    x = np.array([1, 2, 3])
    x_df = assure_pandas_df(x)
    assert x_df.equals(expected_df)
    
def test_assure_pandas_df_df(expected_df_2d):
    x = pd.DataFrame([[1, 2, 3], [2, 3, 4]])
    x_df = assure_pandas_df(x)
    assert x_df.equals(expected_df_2d)
    
def test_assure_pandas_df_types():
    with pytest.raises(TypeError):
        assure_pandas_df("Test")
    with pytest.raises(TypeError):
        assure_pandas_df(5)

def test_check_numeric_dtype_list():
    with pytest.raises(TypeError):
        check_numeric_dtypes(['not numeric', 7, 1.0, True])
    check_numeric_dtypes([1, 2, 3])
    check_numeric_dtypes([1.0, 2.0, 3.0])
    check_numeric_dtypes([False, True, False])
    check_numeric_dtypes([1, True, 7.0])
    
def test_preprocess_labels():
    y1 = pd.Series([1, 0, 1, 0, 1])
    index_1 = np.array([5, 4, 3, 2, 1])

    with pytest.warns(None) as record:
        y1_output = preprocess_labels(y1, y_name='y1', index=index_1, verbose=150)

    pd.testing.assert_series_equal(y1_output, pd.Series([1, 0, 1, 0, 1], index=index_1))
    # Ensure that number of warnings is correct
    assert len(record) == 0

    y2 = [False, False, False, False, False]

    with pytest.warns(None) as record:
        y2_output = preprocess_labels(y2, y_name='y2', verbose=150)

    pd.testing.assert_series_equal(y2_output, pd.Series(y2))
    # Ensure that number of warnings is correct
    assert len(record) == 1

    y3 = np.array([0, 1, 2, 3, 4])
    with pytest.warns(None) as record:
        y3_output = preprocess_labels(y3, y_name='y3', verbose=150)

    pd.testing.assert_series_equal(y3_output, pd.Series(y3))
    # Ensure that number of warnings is correct
    assert len(record) == 1

    y4 = pd.Series(['2', '1', '3', '2', '1'])
    index4 = pd.Index([0, 2, 1, 3, 4])
    with pytest.warns(None) as record:
        y4_output = preprocess_labels(y4, y_name='y4', index=index4, verbose=0)
    print(y4_output)
    print(pd.Series(['2', '3', '1', '2', '1'], index=index4))
    pd.testing.assert_series_equal(y4_output, pd.Series(['2', '3', '1', '2', '1'], index=index4))
    # Ensure that number of warnings is correct
    assert len(record) == 0


def test_preprocess_data():
    X1 = pd.DataFrame({'cat': ['a', 'b', 'c'],
                       'missing': [1, np.nan, 2],
                       'num_1': [1, 2, 3]})

    target_column_names_X1 = ['1', '2', '3']
    X1_expected_output = pd.DataFrame({'1': ['a', 'b', 'c'],
                                       '2': [1, np.nan, 2],
                                       '3': [1, 2, 3]})

    X1_expected_output['1'] = X1_expected_output['1'].astype('category')

    with pytest.warns(None) as record:
        X1_output, output_column_names_X1 = preprocess_data(X1, X_name='X1', column_names=target_column_names_X1,
                                                         verbose=150)

    assert target_column_names_X1 == output_column_names_X1
    pd.testing.assert_frame_equal(X1_output, X1_expected_output)
    # Ensure that number of warnings is correct
    assert len(record) == 2

    X2 = np.array([[1, 3, 2], [1, 2, 2], [1, 2, 3]])

    target_column_names_X1 = [0, 1, 2]
    X2_expected_output = pd.DataFrame(X2, columns=target_column_names_X1)

    with pytest.warns(None) as record:
        X2_output, output_column_names_X2 = preprocess_data(X2, X_name='X2', column_names=None,
                                                         verbose=150)

    assert target_column_names_X1 == output_column_names_X2
    pd.testing.assert_frame_equal(X2_output, X2_expected_output)
    # Ensure that number of warnings is correct
    assert len(record) == 0


# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from .metric import get_metric
from .volatility import (
    BaseVolatilityEstimator,
    TrainTestVolatility,
    BootstrappedVolatility,
    SplitSeedVolatility,
)
from .utils import sample_data, check_sampling_input


__all__ = [
    "get_metric",
    "BaseVolatilityEstimator",
    "TrainTestVolatility",
    "BootstrappedVolatility",
    "SplitSeedVolatility",
    "sample_data",
    "check_sampling_input",
]



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from scipy import stats
from ..utils import assure_numpy_array
from probatus.stat_tests.utils import verbose_p_vals


@verbose_p_vals
def ks(d1, d2, verbose=False):
    """
    Calculates the Kolmogorov-Smirnov statistic on 2 samples. Any binning/bucketing of the distributions/samples
    should be done before passing them to this function.

    Args:
        d1 (np.ndarray or pd.core.series.Series) : first sample.

        d2 (np.ndarray or pd.core.series.Series) : second sample.

        verbose (bool)                           : helpful interpretation msgs printed to stdout (default False).

    Returns:
        (float, float): KS test stat and p-value of rejecting the null hypothesis (that the two distributions are identical)
    """

    # Perform data checks
    d1 = assure_numpy_array(d1)
    d2 = assure_numpy_array(d2)

    # Perform statistical tests
    ks, pvalue = stats.ks_2samp(d1, d2)

    return ks, pvalue



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from abc import ABC, abstractmethod
from probatus.utils import NotFittedError


class BaseFitComputeClass(ABC):
    fitted = False

    def _check_if_fitted(self):
        """
        Checks if object has been fitted. If not, NotFittedError is raised.
        """
        if self.fitted is False:
            raise (
                NotFittedError(
                    "The object has not been fitted. Please run fit() method first"
                )
            )

    @abstractmethod
    def fit(self, *args, **kwargs):
        pass

    @abstractmethod
    def compute(self, *args, **kwargs):
        pass

    @abstractmethod
    def fit_compute(self, *args, **kwargs):
        pass


class BaseFitComputePlotClass(BaseFitComputeClass):
    @abstractmethod
    def plot(self, *args, **kwargs):
        pass



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from sklearn.metrics import get_scorer


def get_scorers(scoring):
    """
    Returns Scorers list based on the provided scoring.

    Args:
        scoring (string, list of strings, probatus.utils.Scorer or list of probatus.utils.Scorers):
            Metrics for which the score is calculated. It can be either a name or list of names metric names and
            needs to be aligned with predefined [classification scorers names in sklearn](https://scikit-learn.org/stable/modules/model_evaluation.html).
            Another option is using probatus.utils.Scorer to define a custom metric.

    Returns:
        (list of probatus.utils.Scorer):
            List of scorers that can be used for scoring models
    """
    scorers = []
    if isinstance(scoring, list):
        for scorer in scoring:
            scorers.append(get_single_scorer(scorer))
    else:
        scorers.append(get_single_scorer(scoring))
    return scorers


def get_single_scorer(scoring):
    """
    Returns single Scorer, based on provided input in scoring argument.

    Args:
        scoring (string or probatus.utils.Scorer, optional):
            Metric for which the model performance is calculated. It can be either a metric name  aligned with
            predefined [classification scorers names in sklearn](https://scikit-learn.org/stable/modules/model_evaluation.html).
            Another option is using probatus.utils.Scorer to define a custom metric.

    Returns:
        (probatus.utils.Scorer):
            Scorer that can be used for scoring models
    """
    if isinstance(scoring, str):
        return Scorer(scoring)
    elif isinstance(scoring, Scorer):
        return scoring
    else:
        raise (
            ValueError(
                "The scoring should contain either strings or probatus.utils.Scorer class"
            )
        )


class Scorer:
    """
    Scores a given machine learning model based on the provided metric name and optionally a custom scoring function.

    Examples:

    ```python
    from probatus.utils import Scorer
    from sklearn.metrics import make_scorer
    from sklearn.datasets import make_classification
    from sklearn.model_selection import train_test_split
    from sklearn.ensemble import RandomForestClassifier
    import pandas as pd

    # Make ROC AUC scorer
    scorer1 = Scorer('roc_auc')

    # Make custom scorer with following function:
    def custom_metric(y_true, y_pred):
          return (y_true == y_pred).sum()
    scorer2 = Scorer('custom_metric', custom_scorer=make_scorer(custom_metric))

    # Prepare two samples
    feature_names = ['f1', 'f2', 'f3', 'f4']
    X, y = make_classification(n_samples=1000, n_features=4, random_state=0)
    X = pd.DataFrame(X, columns=feature_names)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

    # Prepare and fit model. Remember about class_weight="balanced" or an equivalent.
    clf = RandomForestClassifier(class_weight='balanced', n_estimators = 100, max_depth=2, random_state=0)
    clf = clf.fit(X_train, y_train)

    # Score model
    score_test_scorer1 = scorer1.score(clf, X_test, y_test)
    score_test_scorer2 = scorer2.score(clf, X_test, y_test)

    print(f'Test ROC AUC is {score_test_scorer1}, Test {scorer2.metric_name} is {score_test_scorer2}')
    ```
    """

    def __init__(self, metric_name, custom_scorer=None):
        """
        Initializes the class.

        Args:
            metric_name (str):
                Name of the metric used to evaluate the model. If the custom_scorer is not passed, the
                metric name needs to be aligned with [classification scorers names in sklearn](https://scikit-learn.org/stable/modules/model_evaluation.html).


            custom_scorer (sklearn.metrics Scorer callable, optional):
                Callable that can score samples.
        """
        self.metric_name = metric_name
        if custom_scorer is not None:
            self.scorer = custom_scorer
        else:
            self.scorer = get_scorer(self.metric_name)

    def score(self, model, X, y):
        """
        Scores the samples model based on the provided metric name

        Args:
            model (model object):
                Model to be scored.

            X (array-like of shape (n_samples,n_features)):
                Samples on which the model is scored.

            y (array-like of shape (n_samples,)):
                Labels on which the model is scored.

        Returns:
            (float):
                Score returned by the model
        """
        return self.scorer(model, X, y)



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from ..utils import NotFittedError, UnsupportedModelError, BaseFitComputeClass
import numpy as np
import pandas as pd
import copy

from sklearn.cluster import KMeans
from probatus.utils import shap_helpers


def return_confusion_metric(y_true, y_score, normalize=False):
    """
    Computes a confusion metric as absolute difference between the y_true and y_score.
    If normalize eis set to tru, it will normalize y_score to the maximum value in the array
    Args:
        y_true: (np.ndarray or pd.Series) true targets
        y_score: (np.ndarray or pd.Series) model output
        normalize: boolean, normalize or not to the maximum value

    Returns: (np.ndarray or pd.Series) conflusion metric

    """

    if normalize:
        y_score = y_score / y_score.max()

    return np.abs(y_true - y_score)


class BaseInspector(BaseFitComputeClass):
    def __init__(self, algotype, **kwargs):
        self.algotype = algotype
        # TODO fix compilatiopn issue on  for hdbscan
        # if algotype =='dbscan':
        #     self.clusterer = hdbscan.HDBSCAN(prediction_data=True,**kwargs)
        if algotype == "kmeans":
            self.clusterer = KMeans(**kwargs)
        else:
            raise UnsupportedModelError(
                "The algorithm {} is not supported".format(algotype)
            )

    def __repr__(self):
        repr_ = "{},\n\t{}".format(self.__class__.__name__, self.algotype)
        if self.fitted:
            repr_ += "\n\tTotal clusters {}".format(
                np.unique(self.clusterer.labels_).shape[0]
            )
        return repr_

    def fit_clusters(self, X):
        """
        Perform the fit of the clusters with the algorithm specified in the constructor
        Args:
            X: input features

        Returns: cluster labels

        """
        self.clusterer.fit(X)
        self.fitted = True

        return self

    def predict_clusters(self, X):
        if not self.fitted:
            raise NotFittedError("Inspector not fitter. Run .fit()")

        labels = None
        if self.algotype == "kmeans":
            labels = self.clusterer.predict(X)
        if self.algotype == "dbscan":
            raise NotImplementedError(
                "Implementation not finished (note the hdbscan package is not imported yet!)"
            )
            # labels, strengths = hdbscan.approximate_predict(self.clusterer, X)

        return labels

    @staticmethod
    def assert_is_dataframe(df):

        if isinstance(df, pd.DataFrame):
            return df

        elif isinstance(df, np.ndarray) and len(df.shape) == 2:
            return pd.DataFrame(df)

        else:
            raise NotImplementedError(
                "Sorry, X needs to be a pd.DataFrame for for a 2 dimensional numpy array"
            )

    @staticmethod
    def assert_is_series(series, index=None):

        if isinstance(series, pd.Series):
            return series
        elif isinstance(series, pd.DataFrame) and series.shape[1] == 1:
            return pd.Series(series.values.ravel(), index=series.index)
        elif (
            isinstance(series, np.ndarray)
            and len(series.shape) == 1
            and index is not None
        ):
            return pd.Series(series, index=index)
        else:
            raise TypeError(
                "The object should be a pd.Series, a dataframe with one collumn or a 1 dimensional numpy array"
            )


class InspectorShap(BaseInspector):
    """
    Class to perform inspection of the model prediction based on Shapley values.

    It uses the calculated shapley values for the train model to build clusters in the shap space.
    For each cluster, an average confusion, average predicted probability and observed rate of a single class is
    calculated.
    Every sub cluster can be retrieved with the function slice_cluster to perform deeper analysis.

    The original dataframe indexing is used in slicing the dataframe, ensuring easy filtering

    Args:
            model: (obj) pretrained model (with sklearn-like API)
            algotype: (str) clustering algorithm (supported are kmeans and hdbscan)
            confusion_metric: (str) Confusion metric to use:
                - "proba": it will calculate the confusion metric as the absolute value of the target minus the predicted
                           probability. This provides a continuous measure of confusion, where 0 indicated correct predictions
                           and the closer the number is to 1, the higher the confusion
            normalize_probability: (boolean) if true, it will normalize the probabilities to the max value when computing
                the confusion metric
            cluster_probabilities: (boolean) if true, uses the model prediction as an input for the cluster prediction
            **kwargs: keyword arguments for the clustering algorithm

    """

    def __init__(
        self,
        model,
        algotype="kmeans",
        confusion_metric="proba",
        normalize_probability=False,
        cluster_probability=False,
        **kwargs
    ):

        super().__init__(algotype, **kwargs)
        self.model = model
        self.isinspected = False
        self.hasmultiple_dfs = False
        self.normalize_proba = normalize_probability
        self.cluster_probabilities = cluster_probability
        self.agg_summary_df = None
        self.set_names = None
        self.confusion_metric = confusion_metric
        self.cluster_report = None
        self.y = None
        self.predicted_proba = None
        self.X_shap = None
        self.clusters = None
        self.init_eval_set_report_variables()

        if confusion_metric not in ["proba"]:
            # TODO implement the target method
            raise NotImplementedError(
                "confusion metric {} not supported. See docstrings".format(
                    confusion_metric
                )
            )

    def __repr__(self):
        repr_ = "{},\n\t{}".format(self.__class__.__name__, self.algotype)
        if self.fitted:
            repr_ += "\n\tTotal clusters {}".format(
                np.unique(self.clusterer.labels_).shape[0]
            )
        return repr_

    def init_eval_set_report_variables(self):
        self.X_shaps = list()
        self.clusters_list = list()
        self.ys = list()
        self.predicted_probas = list()

    def compute_probabilities(self, X):
        """
        Compute the probabilities for the model using the sklearn API
        Args:
            X: Feature set

        Returns: (np.array) probability

        """

        return self.model.predict_proba(X)[:, 1]

    def fit_clusters(self, X):
        """
        Perform the fit of the clusters with the algorithm specified in the constructor
        Args:
            X: input features


        """
        X = copy.deepcopy(X)

        if self.cluster_probabilities:
            X["probs"] = self.predicted_proba

        return super().fit_clusters(X)

    def predict_clusters(self, X):
        """
        Predicts the clusters of the dataset X
        Args:
            X: features

        Returns: cluster labels

        """
        X = copy.deepcopy(X)

        if self.cluster_probabilities:
            X["probs"] = self.predicted_proba

        return super().predict_clusters(X)

    def fit(self, X, y=None, eval_set=None, sample_names=None, **shap_kwargs):
        """
        Fits and orchestrates the cluster calculations
        Args:
            X: (pd.DataFrame) with the features set used to train the model
            y: (pd.Series, default=None): targets used to train the model
            eval_set: (list, default=None). list of tuples in the shape (X,y) containing evaluation samples, for example
                a test sample, validation sample etc... X corresponds to the feature set of the sample, y corresponds
                to the targets of the samples
            sample_names: (list of strings, default=None): list of suffixed for the samples. If none, it will be labelled with
                sample_{i}, where i corresponds to the index of the sample.
                List length must match that of eval_set
            **shap_kwargs:  kwargs to pass to the Shapley Tree Explained
        """

        self.set_names = sample_names
        if sample_names is not None:
            # Make sure that the amount of eval sets matches the set names
            assert len(eval_set) == len(
                sample_names
            ), "set_names must be the same length as eval_set"

        (
            self.y,
            self.predicted_proba,
            self.X_shap,
            self.clusters,
        ) = self.perform_fit_calc(X=X, y=y, fit_clusters=True, **shap_kwargs)
        if eval_set is not None:
            assert isinstance(eval_set, list), "eval_set needs to be a list"

            self.hasmultiple_dfs = True
            # Reset lists in case inspect run multiple times
            self.init_eval_set_report_variables()

            for X_, y_ in eval_set:
                y_, predicted_proba_, X_shap_, clusters_ = self.perform_fit_calc(
                    X=X_, y=y_, fit_clusters=False, **shap_kwargs
                )

                self.X_shaps.append(X_shap_)
                self.ys.append(y_)
                self.predicted_probas.append(predicted_proba_)
                self.clusters_list.append(clusters_)

        return self

    def perform_fit_calc(self, X, y, fit_clusters=False, **shap_kwargs):
        """
        Performs cluster calculations for a specific X and y
        Args:
            X: pd.DataFrame with the features set used to train the model
            y: pd.Series (default None): targets used to train the model
            fit_clusters: flag indicating whether clustering algorithm should be trained with computed shap values
            **shap_kwargs:  kwargs to pass to the Shapley Tree Explained
        """

        X = self.assert_is_dataframe(X)
        y = self.assert_is_series(y, index=X.index)

        # Compute probabilities for the input X using model
        predicted_proba = pd.Series(
            self.compute_probabilities(X), index=y.index, name="pred_proba"
        )

        # Compute SHAP values and cluster them
        X_shap = shap_helpers.shap_to_df(self.model, X, **shap_kwargs)
        if fit_clusters:
            self.fit_clusters(X_shap)
        clusters = pd.Series(
            self.predict_clusters(X_shap), index=y.index, name="cluster_id"
        )
        return y, predicted_proba, X_shap, clusters

    def _compute_report(self):
        """
        Helper function to compute the report of the ispector - performs aggregations per cluster id

        """

        self.summary_df = self.create_summary_df(
            self.clusters, self.y, self.predicted_proba, normalize=self.normalize_proba
        )
        self.agg_summary_df = self.aggregate_summary_df(self.summary_df)

        if self.hasmultiple_dfs:

            self.summary_dfs = [
                self.create_summary_df(
                    clust, y, pred_proba, normalize=self.normalize_proba
                )
                for clust, y, pred_proba in zip(
                    self.clusters_list, self.ys, self.predicted_probas
                )
            ]

            self.agg_summary_dfs = [
                self.aggregate_summary_df(df) for df in self.summary_dfs
            ]

    def compute(self):
        """
        Calculates a report containing the information per cluster.
        Includes the following:
            - cluster id
            - total number of observations in the cluster
            - total number of target 1 in the cluster
            - target 1 rate (ration of target 1 counts/observations)
            - average predicted probabilitites
            - average confusion

        If multiple eval_sets were passed in the inspect() functions, the output will contain those aggregations as well
        The output names will use the sample names provided in the inspect function. Otherwise they will be labelled by
        the suffix sample_{i}, where i is the index of the sample

        Returns: (pd.DataFrame) with above mentioned aggregations.

        """

        if self.cluster_report is not None:
            return self.cluster_report

        self._compute_report()
        out = copy.deepcopy(self.agg_summary_df)

        if self.hasmultiple_dfs:

            for ix, agg_summary_df in enumerate(self.agg_summary_dfs):
                if self.set_names is None:
                    sample_suffix = "sample_{}".format(ix + 1)
                else:
                    sample_suffix = self.set_names[ix]

                out = pd.merge(
                    out,
                    agg_summary_df,
                    how="left",
                    on="cluster_id",
                    suffixes=("", "_{}".format(sample_suffix)),
                )

        self.cluster_report = out
        return self.cluster_report

    def slice_cluster(
        self,
        cluster_id,
        summary_df=None,
        X_shap=None,
        y=None,
        predicted_proba=None,
        complementary=False,
    ):
        """
        Slices the input dataframes by the cluster.

        Args:
            cluster_id: (int or list for multiple cluster_id) cluster ids to to slice
            summary_df: Optional parameter - the summary_df on which the masking should be performed.
                if not passed the slicing is performed on summary generated by inspect method on X and y
            X_shap: Optional parameter - the SHAP values generated from on X on which the masking should be performed.
                if not passed the slicing is performed on X_shap generated by inspect method on X and y
            y: Optional parameter - the y on which the masking should be performed.
                if not passed the slicing is performed on y passed to inspect
            predicted_proba: Optional parameter - the predicted_proba on which the masking should be performed.
                if not passed the slicing is performed on predicted_proba generated by inspect method on X and y
            complementary: flag that returns the cluster_id if set to False, otherwise the complementary dataframe (ie
                those with ~mask)

        Returns: tuple: Dataframe of sliced shapley values, series of sliced targets, sliced probabilities

        """

        if self.cluster_report is None:
            self.compute()

        # Check if input specified by user, otherwise use the ones from self
        if summary_df is None:
            summary_df = self.summary_df
        if X_shap is None:
            X_shap = self.X_shap
        if y is None:
            y = self.y
        if predicted_proba is None:
            predicted_proba = self.predicted_proba

        mask = self.get_cluster_mask(summary_df, cluster_id)
        if not complementary:
            return X_shap[mask], y[mask], predicted_proba[mask]
        else:
            return X_shap[~mask], y[~mask], predicted_proba[~mask]

    def slice_cluster_eval_set(self, cluster_id, complementary=False):
        """
        Slices the input dataframes passed in the eval_set in the inspect function by the cluster id.

        Args:
            cluster_id: (int or list for multiple cluster_id) cluster ids to to slice
            complementary: flag that returns the cluster_id if set to False, otherwise the complementary dataframe (ie
                those with ~mask)

        Returns: list of tuplse: each element of the list containst
                    Dataframe of sliced shapley values, series of sliced targets, sliced probabilities

        """

        if not self.hasmultiple_dfs:
            raise NotFittedError(
                "You did not fit the eval set. Please add an eval set when calling inspect()"
            )

        output = []
        for X_shap, y, predicted_proba, summary_df in zip(
            self.X_shaps, self.ys, self.predicted_probas, self.summary_dfs
        ):
            output.append(
                self.slice_cluster(
                    cluster_id=cluster_id,
                    summary_df=summary_df,
                    X_shap=X_shap,
                    y=y,
                    predicted_proba=predicted_proba,
                    complementary=complementary,
                )
            )
        return output

    @staticmethod
    def get_cluster_mask(df, cluster_id):
        """
        Returns the mask to filter the cluster id
        Args:
            df: dataframe with 'cluster_id' in it
            cluster_id: int or list of cluster ids to mask

        Returns:

        """

        if not isinstance(cluster_id, list):
            cluster_id = [cluster_id]

        mask = df["cluster_id"].isin(cluster_id)
        return mask

    @staticmethod
    def create_summary_df(cluster, y, probas, normalize=False):
        """
        Creates a summary by concatenating the cluster series, the targets, the probabilities and the measured confusion
        Args:
            cluster: pd.Series od clusters
            y: pd.Series od targets
            probas: pd.Series of predicted probabilities of the model
            normalize: boolean (if the predicted probabilities should be normalized to the max value

        Returns: pd.DataFrame (concatenation of the inputs)

        """

        confusion = return_confusion_metric(y, probas, normalize=normalize).rename(
            "confusion"
        )

        summary = [cluster, y.rename("target"), probas, confusion]

        return pd.concat(summary, axis=1)

    @staticmethod
    def aggregate_summary_df(df):
        """
        Performs the aggregations at the cluster_id level needed to generate the report of the inspection
        Args:
            df: input df to aggregate

        Returns: pd.Dataframe with aggregation results
        """

        out = (
            df.groupby("cluster_id")
            .agg(
                total_label_1=pd.NamedAgg(column="target", aggfunc="sum"),
                total_entries=pd.NamedAgg(column="target", aggfunc="count"),
                label_1_rate=pd.NamedAgg(column="target", aggfunc="mean"),
                average_confusion=pd.NamedAgg(column="confusion", aggfunc="mean"),
                average_pred_proba=pd.NamedAgg(column="pred_proba", aggfunc="mean"),
            )
            .reset_index()
            .rename(columns={"index": "cluster_id"})
            .sort_values(by="cluster_id")
        )

        return out

    def fit_compute(self, X, y=None, eval_set=None, sample_names=None, **shap_kwargs):
        """
        Fits and orchestrates the cluster calculations and returns the computed report

        Args:
            X: (pd.DataFrame) with the features set used to train the model
            y: (pd.Series, default=None): targets used to train the model
            eval_set: (list, default=None). list of tuples in the shape (X,y) containing evaluation samples, for example
                a test sample, validation sample etc... X corresponds to the feature set of the sample, y corresponds
                to the targets of the samples
            sample_names: (list of strings, default=None): list of suffixed for the samples. If none, it will be labelled with
                sample_{i}, where i corresponds to the index of the sample.
                List length must match that of eval_set
            **shap_kwargs:  kwargs to pass to the Shapley Tree Explained

        Returns:
            (pd.DataFrame) Report with aggregations described in compute() method.
        """
        self.fit(X, y, eval_set, sample_names, **shap_kwargs)
        return self.compute()



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from scipy import stats

from ..utils import assure_numpy_array
from probatus.stat_tests.utils import verbose_p_vals


@verbose_p_vals
def es(d1, d2, verbose=False):
    """
    Calculates the Epps-Singleton test statistic on 2 distributions. Can be used on continuous or discrete
    distributions. Any binning/bucketing of the distributions/samples should be done before passing them to this
    function.

    Whereas KS relies on the empirical distribution function, ES is based on the empirical characteristic function
    (Epps & Singleton 1986, Goerg & Kaiser 2009).

    Advantages:

    - Unlike the KS, the ES can be used on both continuous & discrete distributions.

    - ES has higher power (vs KS) in many examples.

    Disadvantages:
    - Not recommended for fewer than 25 observations. Instead, use the Anderson-Darling TS. (However, ES can still be
    used for small samples. A correction factor is applied so that the asymptotic TS distribution more closely follows
    the chi-squared distribution, such that p-values can be computed.)

    Args:
        d1 (np.array or pandas.core.series.Series) : first sample.

        d2 (np.array or pandas.core.series.Series) : second sample.

        verbose (bool) : helpful interpretation msgs printed to stdout (default False).

    Returns:
        (float, float): ES test stat and p-value of rejecting the null hypothesis (that the two distributions are identical)
    """

    d1 = assure_numpy_array(d1)
    d2 = assure_numpy_array(d2)

    es, pvalue = stats.epps_singleton_2samp(d1, d2)

    return es, pvalue



import setuptools
import os

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

base_packages = [
    "scikit-learn>=0.22.2",
    "pandas>=1.0.0",
    "matplotlib>=3.1.1",
    "scipy>=1.4.0",
    "joblib>=0.13.2",
    "tqdm>=4.41.0",
    "shap>=0.38.1",
    "numpy>=1.19.0"
]

dev_dep = [
    "flake8>=3.8.3",
    "black>=19.10b0",
    "pre-commit>=2.5.0",
    "mypy>=0.770",
    "flake8-docstrings>=1.4.0",
    "pytest>=6.0.0",
    "pytest-cov>=2.10.0",
    "pyflakes",
    "seaborn>=0.9.0",
    "joblib>=0.13.2",
    "lightgbm>=3.1.0",
    "jupyter>=1.0.0",
    "tabulate>=0.8.7",
    "nbconvert>=6.0.7",
    "pre-commit>=2.7.1",
]

docs_dep = [
    "mkdocs-material>=6.1.0",
    "mkdocs-git-revision-date-localized-plugin>=0.7.2",
    "mkdocs-git-authors-plugin>=0.3.2",
    "mkdocs-table-reader-plugin>=0.4.1",
    "mkdocs-enumerate-headings-plugin>=0.4.3",
    "mkdocs-awesome-pages-plugin>=2.4.0",
    "mkdocs-minify-plugin>=0.3.0",
    "mknotebooks>=0.6.2",
    "mkdocstrings>=0.13.6",
    "mkdocs-print-site-plugin>=0.8.2",
    "mkdocs-markdownextradata-plugin>=0.1.9",
]

setuptools.setup(
    name="probatus",
    version="1.6.2",
    description="Validation of binary classifiers and data used to develop them",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="ING Bank N.V.",
    author_email="mateusz.garbacz@ing.com",
    license="MIT License",
    packages=setuptools.find_packages(exclude=['tests']),
    python_requires=">=3.6",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=base_packages,
    extras_require={
        "all": base_packages + dev_dep + docs_dep,
    },
    url="https://github.com/ing-bank/probatus",
    zip_safe=False,
)



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import pandas as pd
import numpy as np
from probatus.utils import assure_numpy_array
from probatus.metric_volatility.utils import sample_data
from sklearn.model_selection import train_test_split


def get_metric(
    X,
    y,
    clf,
    test_size,
    split_seed,
    scorers,
    train_sampling_type=None,
    test_sampling_type=None,
    train_sampling_fraction=1,
    test_sampling_fraction=1,
):
    """
    Draws random train/test sample from the data using random seed and calculates metric of interest.

    Args:
        X (np.array or pd.DataFrame):
            Dataset with features.

        y (np.array or pd.Series):
            Target of the prediction.

        clf (model object):
            Binary classification model or pipeline.

        test_size (float):
            Fraction of data used for testing the model.

        split_seed (int):
            Randomized seed used for splitting data.

        scorers (list of Scorers):
            List of Scorer objects used to score the trained model.

        train_sampling_type (str, optional):
            String indicating what type of sampling should be applied on train set:

                - `None`: indicates that no additional sampling is done after splitting data,
                - `'bootstrap'`: indicates that sampling with replacement will be performed on train data,
                - `'subsample'`: indicates that sampling without repetition will be performed  on train data.

        test_sampling_type (str, optional):
            string indicating what type of sampling should be applied on test set:

                - `None`: indicates that no additional sampling is done after splitting data
                - `'bootstrap'`: indicates that sampling with replacement will be performed on test data
                - `'subsample'`: indicates that sampling without repetition will be performed  on test data

        train_sampling_fraction (float, optional):
            Fraction of train data sampled, if sample_train_type is not None. Default value is 1.

        test_sampling_fraction (float, optional):
            Fraction of test data sampled, if sample_test_type is not None. Default value is 1.

    Returns:
        (pd.Dataframe):
            Dataframe with results for a given model trained. Rows indicate the metric measured and columns ther results
    """

    if not (isinstance(X, np.ndarray) or isinstance(X, pd.DataFrame)):
        X = assure_numpy_array(X)
    if not (isinstance(X, np.ndarray) or isinstance(X, pd.Series)):
        y = assure_numpy_array(y)

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=test_size, random_state=split_seed, stratify=y
    )

    # Sample data based on the input arguments
    X_train, y_train = sample_data(
        X=X_train,
        y=y_train,
        sampling_type=train_sampling_type,
        sampling_fraction=train_sampling_fraction,
        dataset_name="train",
    )
    X_test, y_test = sample_data(
        X=X_test,
        y=y_test,
        sampling_type=test_sampling_type,
        sampling_fraction=test_sampling_fraction,
        dataset_name="test",
    )

    clf = clf.fit(X_train, y_train)

    results_columns = ["metric_name", "train_score", "test_score", "delta_score"]
    results = pd.DataFrame([], columns=results_columns)

    for scorer in scorers:
        score_train = scorer.score(clf, X_train, y_train)
        score_test = scorer.score(clf, X_test, y_test)
        score_delta = score_train - score_test

        results = results.append(
            pd.DataFrame(
                [[scorer.metric_name, score_train, score_test, score_delta]],
                columns=results_columns,
            )
        )
    return results



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from probatus.metric_volatility.metric import get_metric
from joblib import Parallel, delayed
from tqdm.auto import tqdm
from probatus.utils import (
    get_scorers,
    assure_list_of_strings,
    assure_list_values_allowed,
    BaseFitComputePlotClass,
    preprocess_data,
    preprocess_labels,
)
from probatus.metric_volatility.utils import check_sampling_input
from probatus.stat_tests import DistributionStatistics
import warnings


class BaseVolatilityEstimator(BaseFitComputePlotClass):
    """
    Base object for estimating volatility estimation. This class is a base class, therefore cannot be used on its
        own. Implements common API that can be used by all subclasses.
    """

    def __init__(
        self,
        clf,
        scoring="roc_auc",
        test_prc=0.25,
        n_jobs=1,
        stats_tests_to_apply=None,
        verbose=0,
        random_state=None,
    ):
        """
        Initializes the class

        Args:
            clf (model object):
                Binary classification model or pipeline.

            scoring (string, list of strings, probatus.utils.Scorer or list of probatus.utils.Scorers, optional):
                Metrics for which the score is calculated. It can be either a name or list of names metric names and
                needs to be aligned with predefined [classification scorers names in sklearn](https://scikit-learn.org/stable/modules/model_evaluation.html).
                Another option is using probatus.utils.Scorer to define a custom metric.

            test_prc (float, optional):
                Percentage of input data used as test. By default 0.25.

            n_jobs (int, optional):
                Number of parallel executions. If -1 use all available cores. By default 1.

            stats_tests_to_apply (None, string or list of strings, optional):
                List of tests to apply. Available options:

                - `'ES'`: Epps-Singleton,
                - `'KS'`: Kolmogorov-Smirnov statistic,
                - `'PSI'`: Population Stability Index,
                - `'SW'`: Shapiro-Wilk based difference statistic,
                - `'AD'`: Anderson-Darling TS.

            verbose (int, optional):
                Controls verbosity of the output:

                - 0 - nether prints nor warnings are shown
                - 1 - 50 - only most important warnings and indication of progress in fitting the object.
                - 51 - 100 - shows other warnings and prints
                - above 100 - presents all prints and all warnings (including SHAP warnings).

            random_state (int, optional):
                Random state set at each round of feature elimination. If it is None, the results will not be
                reproducible and in random search at each iteration a different hyperparameters might be tested. For
                reproducible results set it to integer.
        """
        self.clf = clf
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.test_prc = test_prc
        self.iterations_results = None
        self.report = None
        self.verbose = verbose
        self.allowed_stats_tests = list(
            DistributionStatistics.statistical_test_dict.keys()
        )

        # TODO set reasonable default value for the parameter, to choose the statistical test for the user for different
        #  ways to compute volatility
        if stats_tests_to_apply is not None:
            self.stats_tests_to_apply = assure_list_of_strings(
                stats_tests_to_apply, "stats_tests_to_apply"
            )
            assure_list_values_allowed(
                variable=self.stats_tests_to_apply,
                variable_name="stats_tests_to_apply",
                allowed_values=self.allowed_stats_tests,
            )
        else:
            self.stats_tests_to_apply = []

        self.stats_tests_objects = []
        if len(self.stats_tests_to_apply) > 0:
            if self.verbose > 0:
                warnings.warn(
                    "Computing statistics for distributions is an experimental feature. While using it, keep "
                    "in mind that the samples of metrics might be correlated."
                )
            for test_name in self.stats_tests_to_apply:
                self.stats_tests_objects.append(
                    DistributionStatistics(statistical_test=test_name)
                )

        self.scorers = get_scorers(scoring)

    def fit(self, *args, **kwargs):
        """
        Base fit functionality that should be executed before each fit.

        Returns:
            (BaseVolatilityEstimator):
                Fitted object.
        """

        # Set seed for results reproducibility
        if self.random_state is not None:
            np.random.seed(self.random_state)

        # Initialize the report and results
        self.iterations_results = None
        self.report = None
        self.fitted = True
        return self

    def compute(self, metrics=None):
        """
        Reports the statistics.

        Args:
            metrics (str or list of strings, optional):
                Name or list of names of metrics to be plotted. If not all metrics are presented.

        Returns:
            (pandas.Dataframe):
                Report that contains the evaluation mean and std on train and test sets for each metric.
        """

        self._check_if_fitted()
        if self.report is None:
            raise (
                ValueError(
                    "Report is None, thus it has not been computed by fit method. Please extend the "
                    "BaseVolatilityEstimator class, overwrite fit method, and within fit run compute_report()"
                )
            )

        if metrics is None:
            return self.report
        else:
            if not isinstance(metrics, list):
                metrics = [metrics]
            return self.report.loc[metrics]

    def plot(
        self,
        metrics=None,
        bins=10,
        show=True,
        height_per_subplot=5,
        width_per_subplot=5,
    ):
        """
        Plots distribution of the metric

        Args:
            metrics (str or list of strings, optional):
                Name or list of names of metrics to be plotted. If not all metrics are presented.

            bins (int, optional):
                Number of bins into which histogram is built.

            show (bool, optional):
                If True, the plots are showed to the user, otherwise they are not shown. Not showing plot can be useful,
                when you want to edit the returned axis, before showing it.

            height_per_subplot (int, optional):
                Height of each subplot. Default is 5.

            width_per_subplot (int, optional):
                Width of each subplot. Default is 5.

        Returns
            (list(matplotlib.axes)):
                Axes that include the plot.
        """

        target_report = self.compute(metrics=metrics)

        if target_report.shape[0] >= 1:
            fig, axs = plt.subplots(
                target_report.shape[0],
                2,
                figsize=(
                    width_per_subplot * 2,
                    height_per_subplot * target_report.shape[0],
                ),
            )

            # Enable traversing the axs
            axs = axs.flatten()
            axis_index = 0

            for metric, row in target_report.iterrows():
                train, test, delta = self._get_samples_to_plot(metric_name=metric)

                axs[axis_index].hist(
                    train, alpha=0.5, label="Train {}".format(metric), bins=bins
                )
                axs[axis_index].hist(
                    test, alpha=0.5, label="Test {}".format(metric), bins=bins
                )
                axs[axis_index].set_title("Distributions {}".format(metric))
                axs[axis_index].legend(loc="upper right")

                axs[axis_index + 1].hist(
                    delta, alpha=0.5, label="Delta {}".format(metric), bins=bins
                )
                axs[axis_index + 1].set_title("Distributions delta {}".format(metric))
                axs[axis_index + 1].legend(loc="upper right")

                axis_index += 2

            for ax in axs.flat:
                ax.set(xlabel="{} score".format(metric), ylabel="Results count")

            if show:
                plt.show()
            else:
                plt.close()

            return axs

    def _get_samples_to_plot(self, metric_name):
        """
        Selects samples to be plotted.

        Args:
            metric_name (str):
                Name of metric for which the data should be selected.
        """

        current_metric_results = self.iterations_results[
            self.iterations_results["metric_name"] == metric_name
        ]
        train = current_metric_results["train_score"]
        test = current_metric_results["test_score"]
        delta = current_metric_results["delta_score"]

        return train, test, delta

    def _create_report(self):
        """
        Based on the results for each metric for different sampling, mean and std of distributions of all metrics and
        store them as report.
        """

        unique_metrics = self.iterations_results["metric_name"].unique()

        # Get columns which will be filled
        stats_tests_columns = []
        for stats_tests_object in self.stats_tests_objects:
            stats_tests_columns.append(
                "{} statistic".format(stats_tests_object.statistical_test_name)
            )
            stats_tests_columns.append(
                "{} p-value".format(stats_tests_object.statistical_test_name)
            )
        stats_columns = [
            "train_mean",
            "train_std",
            "test_mean",
            "test_std",
            "delta_mean",
            "delta_std",
        ]
        report_columns = stats_columns + stats_tests_columns

        self.report = pd.DataFrame([], columns=report_columns)

        for metric in unique_metrics:
            metric_iterations_results = self.iterations_results[
                self.iterations_results["metric_name"] == metric
            ]
            metrics = self._compute_mean_std_from_runs(metric_iterations_results)
            stats_tests_values = self._compute_stats_tests_values(
                metric_iterations_results
            )
            metric_row = pd.DataFrame(
                [metrics + stats_tests_values], columns=report_columns, index=[metric]
            )
            self.report = self.report.append(metric_row)

    def _compute_mean_std_from_runs(self, metric_iterations_results):
        """
        Compute mean and std of results.

        Args:
            metric_iterations_results (pandas.DataFrame):
                Scores for a single metric for each iteration.

        Returns:
            (list):
                List containing mean and std of train, test and deltas.
        """
        train_mean_score = np.mean(metric_iterations_results["train_score"])
        test_mean_score = np.mean(metric_iterations_results["test_score"])
        delta_mean_score = np.mean(metric_iterations_results["delta_score"])
        train_std_score = np.std(metric_iterations_results["train_score"])
        test_std_score = np.std(metric_iterations_results["test_score"])
        delta_std_score = np.std(metric_iterations_results["delta_score"])
        return [
            train_mean_score,
            train_std_score,
            test_mean_score,
            test_std_score,
            delta_mean_score,
            delta_std_score,
        ]

    def _compute_stats_tests_values(self, metric_iterations_results):
        """
        Compute statistics and p-values of specified tests.

        Args:
            metric_iterations_results (pandas.DataFrame):
                Scores for a single metric for each iteration.

        Returns:
            (list):
                List containing statistics and p-values of distributions.
        """
        statistics = []
        for stats_test in self.stats_tests_objects:
            stats, p_value = stats_test.compute(
                metric_iterations_results["test_score"],
                metric_iterations_results["train_score"],
            )
            statistics += [stats, p_value]
        return statistics

    def fit_compute(self, *args, **kwargs):
        """
        Runs trains and evaluates a number of models on train and test sets extracted using different random seeds.
            Reports the statistics of the selected metric.

        Takes as arguments the same parameters as fit() method.

        Returns:
            (pandas.Dataframe):
                Report that contains the evaluation mean and std on train and test sets for each metric.
        """

        self.fit(*args, **kwargs)
        return self.compute()


class TrainTestVolatility(BaseVolatilityEstimator):
    """
    Estimation of volatility of metrics. The estimation is done by splitting the data into train and test multiple times
        and training and scoring a model based on these metrics. The class allows for choosing whether at each iteration
        the train test split should be the same or different, whether and how the train and test sets should be sampled.

    Examples:
    ```python
    from sklearn.datasets import make_classification
    from sklearn.ensemble import RandomForestClassifier
    from probatus.metric_volatility import TrainTestVolatility
    X, y = make_classification(n_features=4)
    clf = RandomForestClassifier()
    volatility = TrainTestVolatility(clf, iterations=10 , test_prc = 0.5)
    volatility_report = volatility.fit_compute(X, y)
    volatility.plot()
    ```

    <img src="../img/metric_volatility_train_test.png" width="500" />
    """

    def __init__(
        self,
        clf,
        iterations=1000,
        scoring="roc_auc",
        sample_train_test_split_seed=True,
        train_sampling_type=None,
        test_sampling_type=None,
        train_sampling_fraction=1,
        test_sampling_fraction=1,
        test_prc=0.25,
        n_jobs=1,
        stats_tests_to_apply=None,
        verbose=0,
        random_state=None,
    ):

        """
        Initializes the class.

        Args:
            clf (model object):
                Binary classification model or pipeline.

            iterations (int, optional):
                Number of iterations in seed bootstrapping. By default 1000.

            scoring (string, list of strings, probatus.utils.Scorer or list of probatus.utils.Scorers, optional):
                Metrics for which the score is calculated. It can be either a name or list of names metric names and
                needs to be aligned with predefined [classification scorers names in sklearn](https://scikit-learn.org/stable/modules/model_evaluation.html).
                Another option is using probatus.utils.Scorer to define a custom metric.

            sample_train_test_split_seed (bool, optional):
                Flag indicating whether each train test split should be done
                randomly or measurement should be done for single split. Default is True, which indicates that each.
                iteration is performed on a random train test split. If the value is False, the random_seed for the
                split is set to train_test_split_seed.

            train_sampling_type (str, optional):
                String indicating what type of sampling should be applied on train set:

                - `None` indicates that no additional sampling is done after splitting data,
                - `'bootstrap'` indicates that sampling with replacement will be performed on train data,
                - `'subsample'` indicates that sampling without repetition will be performed  on train data.

            test_sampling_type (str, optional):
                String indicating what type of sampling should be applied on test set:

                - `None` indicates that no additional sampling is done after splitting data,
                - `'bootstrap'` indicates that sampling with replacement will be performed on test data,
                - `'subsample'` indicates that sampling without repetition will be performed  on test data.

            train_sampling_fraction (float, optional):
                Fraction of train data sampled, if sample_train_type is not None.
                Default value is 1.

            test_sampling_fraction (float, optional):
                Fraction of test data sampled, if sample_test_type is not None. Default value is 1.

            test_prc (float, optional):
                Percentage of input data used as test. By default 0.25.

            n_jobs (int, optional):
                Number of parallel executions. If -1 use all available cores. By default 1.

            stats_tests_to_apply (None, string or list of strings, optional):
                List of tests to apply, default is None. Available options:

                - `'ES'`: Epps-Singleton,
                - `'KS'`: Kolmogorov-Smirnov statistic,
                - `'PSI'`: Population Stability Index,
                - `'SW'`: Shapiro-Wilk based difference statistic,
                - `'AD'`: Anderson-Darling TS.

            verbose (int, optional):
                Controls verbosity of the output:

                - 0 - nether prints nor warnings are shown
                - 1 - 50 - only most important warnings
                - 51 - 100 - shows other warnings and prints
                - above 100 - presents all prints and all warnings (including SHAP warnings).

            random_state (int, optional):
                Random state set at each round of feature elimination. If it is None, the results will not be
                reproducible and in random search at each iteration a different hyperparameters might be tested. For
                reproducible results set it to integer.
        """
        super().__init__(
            clf=clf,
            scoring=scoring,
            test_prc=test_prc,
            n_jobs=n_jobs,
            stats_tests_to_apply=stats_tests_to_apply,
            verbose=verbose,
            random_state=random_state,
        )
        self.iterations = iterations
        self.train_sampling_type = train_sampling_type
        self.test_sampling_type = test_sampling_type
        self.sample_train_test_split_seed = sample_train_test_split_seed
        self.train_sampling_fraction = train_sampling_fraction
        self.test_sampling_fraction = test_sampling_fraction

        check_sampling_input(train_sampling_type, train_sampling_fraction, "train")
        check_sampling_input(test_sampling_type, test_sampling_fraction, "test")

    def fit(self, X, y, column_names=None):
        """
        Bootstraps a number of random seeds, then splits the data based on the sampled seeds and estimates performance
            of the model based on the split data.

        Args:
            X (pandas.DataFrame or numpy.ndarray):
                Array with samples and features.

            y (pandas.Series or numpy.ndarray):
                Array with targets.

            column_names (list of str, optional):
                List of feature names of the provided samples. If provided it will be used to overwrite the existing
                feature names. If not provided the existing feature names are used or default feature names are
                generated.
        Returns:
            (TrainTestVolatility):
                Fitted object.
        """
        super().fit()

        self.X, self.column_names = preprocess_data(
            X, X_name="X", column_names=column_names, verbose=self.verbose
        )
        self.y = preprocess_labels(
            y, y_name="y", index=self.X.index, verbose=self.verbose
        )

        if self.sample_train_test_split_seed:
            random_seeds = np.random.random_integers(0, 999999, self.iterations)
        else:
            random_seeds = (np.ones(self.iterations)).astype(int)
            if self.random_state:
                random_seeds = random_seeds * self.random_state

        if self.verbose > 0:
            random_seeds = tqdm(random_seeds)

        results_per_iteration = Parallel(n_jobs=self.n_jobs)(
            delayed(get_metric)(
                X=self.X,
                y=self.y,
                clf=self.clf,
                test_size=self.test_prc,
                split_seed=split_seed,
                scorers=self.scorers,
                train_sampling_type=self.train_sampling_type,
                test_sampling_type=self.test_sampling_type,
                train_sampling_fraction=self.train_sampling_fraction,
                test_sampling_fraction=self.test_sampling_fraction,
            )
            for split_seed in random_seeds
        )

        self.iterations_results = pd.concat(results_per_iteration, ignore_index=True)

        self._create_report()
        return self


class SplitSeedVolatility(TrainTestVolatility):
    """
    Estimation of volatility of metrics depending on the seed used to split the data. At every iteration it splits the
        data into train and test set using a different stratified split and volatility of the metrics is calculated.

    Examples:
    ```python
    from sklearn.datasets import make_classification
    from sklearn.ensemble import RandomForestClassifier
    from probatus.metric_volatility import SplitSeedVolatility
    X, y = make_classification(n_features=4)
    clf = RandomForestClassifier()
    volatility = SplitSeedVolatility(clf, iterations=10 , test_prc = 0.5)
    volatility_report = volatility.fit_compute(X, y)
    volatility.plot()
    ```
    <img src="../img/metric_volatility_split_seed.png" width="500" />

    """

    def __init__(
        self,
        clf,
        iterations=1000,
        scoring="roc_auc",
        test_prc=0.25,
        n_jobs=1,
        stats_tests_to_apply=None,
        verbose=0,
        random_state=None,
    ):
        """
        Initializes the class

        Args:
            clf (model object):
                Binary classification model or pipeline.

            iterations (int, optional):
                Number of iterations in seed bootstrapping. By default 1000.

            scoring (string, list of strings, probatus.utils.Scorer or list of probatus.utils.Scorers, optional):
                Metrics for which the score is calculated. It can be either a name or list of names metric names and
                needs to be aligned with predefined [classification scorers names in sklearn](https://scikit-learn.org/stable/modules/model_evaluation.html).
                Another option is using probatus.utils.Scorer to define a custom metric.

            test_prc (float, optional):
                Percentage of input data used as test. By default 0.25.

            n_jobs (int, optional):
                Number of parallel executions. If -1 use all available cores. By default 1.

            stats_tests_to_apply (None, string or list of strings, optional):
                List of tests to apply, default is None. Available options:

                - `'ES'`: Epps-Singleton,
                - `'KS'`: Kolmogorov-Smirnov statistic,
                - `'PSI'`: Population Stability Index,
                - `'SW'`: Shapiro-Wilk based difference statistic,
                - `'AD'`: Anderson-Darling TS.

            verbose (int, optional):
                Controls verbosity of the output:

                - 0 - nether prints nor warnings are shown
                - 1 - 50 - only most important warnings
                - 51 - 100 - shows other warnings and prints
                - above 100 - presents all prints and all warnings (including SHAP warnings).

            random_state (int, optional):
                Random state set at each round of feature elimination. If it is None, the results will not be
                reproducible and in random search at each iteration a different hyperparameters might be tested. For
                reproducible results set it to integer.
        """
        super().__init__(
            clf=clf,
            sample_train_test_split_seed=True,
            train_sampling_type=None,
            test_sampling_type=None,
            train_sampling_fraction=1,
            test_sampling_fraction=1,
            iterations=iterations,
            scoring=scoring,
            test_prc=test_prc,
            n_jobs=n_jobs,
            stats_tests_to_apply=stats_tests_to_apply,
            verbose=verbose,
            random_state=random_state,
        )


class BootstrappedVolatility(TrainTestVolatility):
    """
    Estimation of volatility of metrics by bootstrapping both train and test set. By default at every iteration the
        train test split is the same. The test shows volatility of metric with regards to sampling different rows from
        static train and test sets.

    Examples:
    ```python
    from sklearn.datasets import make_classification
    from sklearn.ensemble import RandomForestClassifier
    from probatus.metric_volatility import BootstrappedVolatility
    X, y = make_classification(n_features=4)
    clf = RandomForestClassifier()
    volatility = BootstrappedVolatility(clf, iterations=10 , test_prc = 0.5)
    volatility_report = volatility.fit_compute(X, y)
    volatility.plot()
    ```
    <img src="../img/metric_volatility_bootstrapped.png" width="500" />
    """

    def __init__(
        self,
        clf,
        iterations=1000,
        scoring="roc_auc",
        train_sampling_fraction=1,
        test_sampling_fraction=1,
        test_prc=0.25,
        n_jobs=1,
        stats_tests_to_apply=None,
        verbose=0,
        random_state=None,
    ):
        """
        Initializes the class.

        Args:
            clf (model object):
                Binary classification model or pipeline.

            iterations (int, optional):
                Number of iterations in seed bootstrapping. By default 1000.

            scoring (string, list of strings, probatus.utils.Scorer or list of probatus.utils.Scorers, optional):
                Metrics for which the score is calculated. It can be either a name or list of names metric names and
                needs to be aligned with predefined [classification scorers names in sklearn](https://scikit-learn.org/stable/modules/model_evaluation.html).
                Another option is using probatus.utils.Scorer to define a custom metric.

            train_sampling_fraction (float, optional):
                Fraction of train data sampled, if sample_train_type is not None. Default value is 1.

            test_sampling_fraction (float, optional):
                Fraction of test data sampled, if sample_test_type is not None. Default value is 1.

            test_prc (float, optional):
                Percentage of input data used as test. By default 0.25.

            n_jobs (int, optional):
                Number of parallel executions. If -1 use all available cores. By default 1.

            stats_tests_to_apply (None, string or list of strings, optional):
                List of tests to apply, default is None. Available options:

                - `'ES'`: Epps-Singleton,
                - `'KS'`: Kolmogorov-Smirnov statistic,
                - `'PSI'`: Population Stability Index,
                - `'SW'`: Shapiro-Wilk based difference statistic,
                - `'AD'`: Anderson-Darling TS.

            verbose (int, optional):
                Controls verbosity of the output:

                - 0 - nether prints nor warnings are shown
                - 1 - 50 - only most important warnings
                - 51 - 100 - shows other warnings and prints
                - above 100 - presents all prints and all warnings (including SHAP warnings).

            random_state (int, optional):
                Random state set at each round of feature elimination. If it is None, the results will not be
                reproducible and in random search at each iteration a different hyperparameters might be tested. For
                reproducible results set it to integer.
        """
        super().__init__(
            clf=clf,
            sample_train_test_split_seed=False,
            train_sampling_type="bootstrap",
            test_sampling_type="bootstrap",
            iterations=iterations,
            scoring=scoring,
            train_sampling_fraction=train_sampling_fraction,
            test_sampling_fraction=test_sampling_fraction,
            test_prc=test_prc,
            n_jobs=n_jobs,
            stats_tests_to_apply=stats_tests_to_apply,
            verbose=verbose,
            random_state=random_state,
        )



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import matplotlib.pyplot as plt
from matplotlib.pyplot import figure


def plot_distributions_of_feature(
    feature_distributions,
    feature_name=None,
    sample_names=None,
    plot_bw_method=0.05,
    plot_perc_outliers_removed=0.01,
    plot_figsize=(15, 6),
):
    """
    This function plots multiple distributions of the same feature. It is e.g. useful to compare
    distribution between train and test.

    For categorical feature the plot bar is plotted, and for numeric the density plot.

    Args:
        feature_distributions (list of pd.Series): List of distributions of the same feature, e.g. values of feature 'f1'
        for Train, Validation and Test.

        feature_name (Optional, str): Name of the feature plotted.

        sample_names (Optional, list of str): List of names of samples e.g. ['Train', 'Validation', 'Test'].

        plot_bw_method (Optional, float): Estimator bandwidth in density plot.

        plot_perc_outliers_removed (Optional, float): Percentage of outliers removed from each side before plotting.

        plot_figsize (Optional, tuple): Size of the figure.

    """
    figure(figsize=plot_figsize)

    if feature_name is None:
        feature_name = feature_distributions[0].name

    if sample_names is None:
        sample_names = [f"sample_{i}" for i in range(len(feature_distributions))]

    if feature_distributions[0].dtype.name == "category":
        data_dict = {}

        for feature_distribution_index in range(len(feature_distributions)):
            data_dict[sample_names[feature_distribution_index]] = feature_distributions[
                feature_distribution_index
            ].value_counts(normalize=True)

        plt.ylabel("Relative frequencies of values in feature.")
    else:
        for feature_distribution_index in range(len(feature_distributions)):
            current_feature = feature_distributions[feature_distribution_index]

            # Remove outliers in each feature
            current_feature = current_feature[
                current_feature.between(
                    current_feature.quantile(0 + plot_perc_outliers_removed),
                    current_feature.quantile(1 - plot_perc_outliers_removed),
                )
            ]

            # Plot density plot
            current_feature.plot.density(bw_method=plot_bw_method)

    plt.title(f"Distribution of {feature_name}")
    plt.xlabel("Feature Distribution")
    plt.legend(sample_names, loc="upper right")
    plt.show()



from sklearn.tree import DecisionTreeClassifier
from sklearn.linear_model import LogisticRegression
import pytest
import numpy as np
import pandas as pd
from probatus.interpret import ShapModelInterpreter
from unittest.mock import patch
import os


@pytest.fixture(scope='function')
def X_train():
    return pd.DataFrame({'col_1': [1, 1, 1, 1],
                         'col_2': [0, 0, 0, 0],
                         'col_3': [1, 0, 1, 0]}, index=[1, 2, 3, 4])


@pytest.fixture(scope='function')
def y_train():
    return pd.Series([1, 0, 1, 0], index=[1, 2, 3, 4])


@pytest.fixture(scope='function')
def X_test():
    return pd.DataFrame({'col_1': [1, 1, 1, 1],
                         'col_2': [0, 0, 0, 0],
                         'col_3': [1, 0, 1, 0]}, index=[5, 6, 7, 8])


@pytest.fixture(scope='function')
def y_test():
    return pd.Series([0, 0, 1, 0], index=[5, 6, 7, 8])


@pytest.fixture(scope='function')
def fitted_tree(X_train, y_train):
    return DecisionTreeClassifier(max_depth=1, random_state=1).fit(X_train, y_train)

@pytest.fixture(scope='function')
def fitted_lin(X_train, y_train):
    return LogisticRegression(random_state=1).fit(X_train, y_train)

@pytest.fixture(scope='function')
def expected_feature_importance():
    return pd.DataFrame({
        'mean_abs_shap_value_test': [0.5, 0., 0.],
        'mean_abs_shap_value_train': [0.5, 0., 0.],
        'mean_shap_value_test': [-0.5, 0., 0.],
        'mean_shap_value_train': [-0.5, 0., 0.]}, index=['col_3', 'col_1', 'col_2'])

@pytest.fixture(scope='function')
def expected_feature_importance_lin_models():
    return pd.DataFrame({
        'mean_abs_shap_value_test': [0.4, 0., 0.],
        'mean_abs_shap_value_train': [0.4, 0., 0.],
        'mean_shap_value_test': [-0.4, 0., 0.],
        'mean_shap_value_train': [-0.4, 0., 0.]}, index=['col_3', 'col_1', 'col_2'])

def test_shap_interpret(fitted_tree, X_train, y_train, X_test, y_test, expected_feature_importance):
    class_names = ['neg', 'pos']

    shap_interpret = ShapModelInterpreter(fitted_tree)
    shap_interpret.fit(X_train, X_test, y_train, y_test, class_names=class_names)

    # Check parameters
    assert shap_interpret.fitted == True
    shap_interpret._check_if_fitted

    assert shap_interpret.class_names == class_names
    assert shap_interpret.train_score == 1
    assert shap_interpret.test_score == pytest.approx(0.833, 0.01)

    # Check expected shap values
    assert (np.mean(np.abs(shap_interpret.shap_values_test), axis=0) == [0, 0, 0.5]).all()
    assert (np.mean(np.abs(shap_interpret.shap_values_train), axis=0) == [0, 0, 0.5]).all()

    importance_df, train_auc, test_auc = shap_interpret.compute(return_scores=True)

    pd.testing.assert_frame_equal(expected_feature_importance, importance_df)
    assert train_auc == 1
    assert test_auc == pytest.approx(0.833, 0.01)

    # Check if plots work for such dataset
    with patch('matplotlib.pyplot.figure') as mock_plt:
        with patch('shap.plots._waterfall.waterfall_legacy'):
            ax1 = shap_interpret.plot('importance', target_set='test', show=False)
            ax2 = shap_interpret.plot('summary', target_set='test', show=False)
            ax3 = shap_interpret.plot('dependence', target_columns='col_3', target_set='test', show=False)
            ax4 = shap_interpret.plot('sample', samples_index=X_test.index.tolist()[0:2], target_set='test', show=False)
            ax5 = shap_interpret.plot('importance', target_set='train')
            ax6 = shap_interpret.plot('summary', target_set='train')
            ax7 = shap_interpret.plot('dependence', target_columns='col_3', target_set='train')
            ax8 = shap_interpret.plot('sample', samples_index=X_train.index.tolist()[0:2], target_set='train')
    assert not(isinstance(ax1, list))
    assert not(isinstance(ax2, list))
    assert isinstance(ax3, list) and len(ax4) == 2
    assert isinstance(ax4, list) and len(ax4) == 2
    assert not(isinstance(ax5, list))
    assert not(isinstance(ax6, list))
    assert isinstance(ax7, list) and len(ax7) == 2
    assert isinstance(ax8, list) and len(ax8) == 2

def test_shap_interpret_lin_models(fitted_lin, X_train, y_train, X_test, y_test, expected_feature_importance_lin_models):
    class_names = ['neg', 'pos']

    shap_interpret = ShapModelInterpreter(fitted_lin)
    shap_interpret.fit(X_train, X_test, y_train, y_test, class_names=class_names)

    # Check parameters
    assert shap_interpret.fitted == True
    shap_interpret._check_if_fitted

    assert shap_interpret.class_names == class_names
    assert shap_interpret.train_score == 1
    assert shap_interpret.test_score == pytest.approx(0.833, 0.01)

    # Check expected shap values
    assert (np.round(np.mean(np.abs(shap_interpret.shap_values_test), axis=0),2) == [0, 0, 0.4]).all()
    assert (np.round(np.mean(np.abs(shap_interpret.shap_values_train), axis=0),2) == [0, 0, 0.4]).all()

    importance_df, train_auc, test_auc = shap_interpret.compute(return_scores=True)
    importance_df=importance_df.round(2)
    
    pd.testing.assert_frame_equal(expected_feature_importance_lin_models, importance_df)
    assert train_auc == 1
    assert test_auc == pytest.approx(0.833, 0.01)

    # Check if plots work for such dataset
    with patch('matplotlib.pyplot.figure') as mock_plt:
        with patch('shap.plots._waterfall.waterfall_legacy'):
            ax1 = shap_interpret.plot('importance', target_set='test', show=False)
            ax2 = shap_interpret.plot('summary', target_set='test', show=False)
            ax3 = shap_interpret.plot('dependence', target_columns='col_3', target_set='test', show=False)
            ax4 = shap_interpret.plot('sample', samples_index=X_test.index.tolist()[0:2], target_set='test', show=False)
            ax5 = shap_interpret.plot('importance', target_set='train')
            ax6 = shap_interpret.plot('summary', target_set='train')
            ax7 = shap_interpret.plot('dependence', target_columns='col_3', target_set='train')
            ax8 = shap_interpret.plot('sample', samples_index=X_train.index.tolist()[0:2], target_set='train')
    assert not(isinstance(ax1, list))
    assert not(isinstance(ax2, list))
    assert isinstance(ax3, list) and len(ax4) == 2
    assert isinstance(ax4, list) and len(ax4) == 2
    assert not(isinstance(ax5, list))
    assert not(isinstance(ax6, list))
    assert isinstance(ax7, list) and len(ax7) == 2
    assert isinstance(ax8, list) and len(ax8) == 2


def test_shap_interpret_fit_compute_lin_models(fitted_lin, X_train, y_train, X_test, y_test, expected_feature_importance_lin_models):
    class_names = ['neg', 'pos']

    shap_interpret = ShapModelInterpreter(fitted_lin)
    importance_df = shap_interpret.fit_compute(X_train, X_test, y_train, y_test, class_names=class_names)
    importance_df=importance_df.round(2)

    # Check parameters
    assert shap_interpret.fitted == True
    shap_interpret._check_if_fitted

    assert shap_interpret.class_names == class_names
    assert shap_interpret.train_score == 1

    assert shap_interpret.test_score == pytest.approx(0.833, 0.01)

    # Check expected shap values
    assert (np.round(np.mean(np.abs(shap_interpret.shap_values_test), axis=0),2) == [0, 0, 0.4]).all()
    assert (np.round(np.mean(np.abs(shap_interpret.shap_values_train), axis=0),2) == [0, 0, 0.4]).all()

    pd.testing.assert_frame_equal(expected_feature_importance_lin_models, importance_df)

def test_shap_interpret_fit_compute(fitted_tree, X_train, y_train, X_test, y_test, expected_feature_importance):
    class_names = ['neg', 'pos']

    shap_interpret = ShapModelInterpreter(fitted_tree)
    importance_df = shap_interpret.fit_compute(X_train, X_test, y_train, y_test, class_names=class_names)

    # Check parameters
    assert shap_interpret.fitted == True
    shap_interpret._check_if_fitted

    assert shap_interpret.class_names == class_names
    assert shap_interpret.train_score == 1
    assert shap_interpret.test_score == pytest.approx(0.833, 0.01)

    # Check expected shap values
    assert (np.mean(np.abs(shap_interpret.shap_values_test), axis=0) == [0, 0, 0.5]).all()
    assert (np.mean(np.abs(shap_interpret.shap_values_train), axis=0) == [0, 0, 0.5]).all()

    pd.testing.assert_frame_equal(expected_feature_importance, importance_df)

@pytest.mark.skipif(os.environ.get("SKIP_LIGHTGBM") == 'true', reason="LightGBM tests disabled")
def test_shap_interpret_complex_data(complex_data_split, complex_fitted_lightgbm):
    class_names = ['neg', 'pos']
    X_train, X_test, y_train, y_test = complex_data_split

    shap_interpret = ShapModelInterpreter(complex_fitted_lightgbm, verbose=50)
    with pytest.warns(None) as record:
        importance_df = shap_interpret.fit_compute(X_train, X_test, y_train, y_test, class_names=class_names)

    # Check parameters
    assert shap_interpret.fitted == True
    shap_interpret._check_if_fitted

    assert shap_interpret.class_names == class_names
    assert len(record) > 4

    assert importance_df.shape[0] == X_train.shape[1]

    # Check if plots work for such dataset
    with patch('matplotlib.pyplot.figure') as mock_plt:
        with patch('shap.plots._waterfall.waterfall_legacy'):
            ax1 = shap_interpret.plot('importance', target_set='test')
            ax2 = shap_interpret.plot('summary', target_set='test')
            ax3 = shap_interpret.plot('dependence', target_columns='f2_missing', target_set='test')
            ax4 = shap_interpret.plot('sample', samples_index=X_test.index.tolist()[0:2], target_set='test')
            ax5 = shap_interpret.plot('importance', target_set='train')
            ax6 = shap_interpret.plot('summary', target_set='train')
            ax7 = shap_interpret.plot('dependence', target_columns='f2_missing', target_set='train')
            ax8 = shap_interpret.plot('sample', samples_index=X_train.index.tolist()[0:2], target_set='train')
    assert not(isinstance(ax1, list))
    assert not(isinstance(ax2, list))
    assert isinstance(ax3, list) and len(ax4) == 2
    assert isinstance(ax4, list) and len(ax4) == 2
    assert not(isinstance(ax5, list))
    assert not(isinstance(ax6, list))
    assert isinstance(ax7, list) and len(ax7) == 2
    assert isinstance(ax8, list) and len(ax8) == 2